<?php
/**
 * EGroupware API - Mime Magic
 *
 * Originally taken from the Horde Framework http://horde.org
 *
 * Copyright 1999-2003 Anil Madhavapeddy <anil@recoil.org>
 * Copyright 2002-2003 Michael Slusarz <slusarz@bigworm.colorado.edu>
 * Copyright 2003 Free Software Foundation, Inc.
 *
 * Generated: 12/04/14 00:15:12 by slusarz on bigworm.curecanti.org
 *
 * Ported to phpGroupWare by Dave Hall - dave.hall@mbox.com.au
 * Note: this class was relicensed as GPL by Dave Hall - all mods GPL
 *
 * @link http://www.egroupware.org
 * @package API
 * @license http://opensource.org/licenses/gpl-license.php GPL - GNU General Public License
 * @version $Id$
 */

namespace EGroupware\Api;

/**
 * Mime Magic class
 *
 * All methods can be called static, no need to instanciate the class.
 */
class MimeMagic
{
	/**
	 * Get a user friendly label for a mime type: e.g. "PDF file (application/pdf)"
	 *
	 * @param string $_mime
	 * @return string
	 */
	public static function mime2label($_mime)
	{
		$mime = strtolower($_mime);

		if ($mime == Vfs::DIR_MIME_TYPE)
		{
			return lang('Directory');
		}
		elseif (!($ext = self::mime2ext($mime)))
		{
			return $mime;
		}
		return lang('%1 file',strtoupper($ext)).' ('.$mime.')';
	}

	/**
	 * Convert a file extension to a MIME type
	 *
	 * This is the simplest MIME type guessing function - rough but fast.
	 * If the MIME type is not found then 'application/octet-stream'
	 * is returned.
	 *
	 * @param string $_ext  The file extension to be mapped to a MIME type.
	 * @return string  The MIME type of the file extension.
	 */
	public static function ext2mime($_ext)
	{
		if (empty($_ext))
		{
			return 'text/plain';//assume no extension is a text file
		}
		$ext = strtolower($_ext);
		if (!array_key_exists($ext, self::$mime_extension_map))
		{
			return 'application/octet-stream';
		}
		return self::$mime_extension_map[$ext];
	}

	/**
	 * Convert a filename to a MIME type, based on the
	 * global and application specific config files.
	 *
	 * Unlike ext2mime, this function will return
	 * 'application/octet-stream' for any unknown or empty extension
	 *
	 * @param string $filename  The filename to be mapped to a MIME type.
	 * @return string  The MIME type of the filename.
	 * @author skwashd - changed it to make it work with file.tar.gz etc
	 */
	public static function filename2mime($filename)
	{
		$fn_parts = explode('.', $filename);
		if (is_array($fn_parts))
		{
			return self::ext2mime(array_pop($fn_parts));
		}
		return 'application/octet-stream';
	}

	/**
	 * Convert a MIME type to a file extension, based
	 * on the global Horde and application specific config files.
	 *
	 * If we cannot map the type to a file extension, we return false.
	 *
	 * @param string $_type  The MIME type to be mapped to a file extension.
	 * @return string  The file extension of the MIME type.
	 */
	public static function mime2ext($_type)
	{
		$type = strtolower($_type);
		if (isset(self::$mime_alias_map[$type])) $type = self::$mime_alias_map[$type];
		$key = array_search($type, self::$mime_extension_map);
		if (empty($type) || $key === false)
		{
			return false;
		}
		return $key;
	}

	/**
	 * Convert a MIME type to all known file extensions
	 *
	 * If we cannot map the type to any file extension, we return an empty array.
	 *
	 * @param string $_type  The MIME type to be mapped to a file extension.
	 * @return array with possible file extensions
	 */
	public static function mime2extensions($_type)
	{
		$type = strtolower($_type);
		if (isset(self::$mime_alias_map[$type])) $type = self::$mime_alias_map[$type];

		return array_keys(array_filter(self::$mime_extension_map, static function($mime) use ($type)
		{
			return $mime == $type;
		}));
	}

	/**
	 * Fix old / aliased mime-types by returning valid/default mime-type
	 *
	 * @param string $_alias
	 * @return string new type, of $alias, if there is no alias defined for it
	 */
	public static function fix_mime_type($_alias)
	{
		$alias = strtolower($_alias);

		if (isset(self::$mime_alias_map[$alias]))
		{
			return self::$mime_alias_map[$alias];
		}
		return $alias;
	}

	/**
	 * Uses variants of the UNIX "file" command to attempt to determine the
	 * MIME type of an unknown file.
	 *
	 * @param string $path filename (including full path) of file to analyze
	 *
	 * @return string  The MIME type of the file.  Returns false if either
	 *                 the file type isn't recognized or the file command is
	 *                 not available.
	 */
	public static function analyze_file($path)
	{
		// If the PHP Mimetype extension is available, use that.
		if (function_exists('mime_content_type'))
		{
			return mime_content_type($path);
		}
		else
		{
			// Use a built-in magic file.
			if (!($fp = @fopen($path, 'rb')))
			{
				return false;
			}
			foreach (self::$mime_magic_file as $offset => $odata)
			{
				foreach ($odata as $length => $ldata)
				{
					@fseek($fp, $offset, SEEK_SET);
					$lookup = @fread($fp, $length);
					if (!empty($ldata[$lookup]))
					{
						fclose($fp);
						return $ldata[$lookup];
					}
				}
			}
			fclose($fp);
		}
		return false;
	}

	/**
	 * Instead of using an existing file a chunk of data is used for
	 * testing.  Best to handle the file creation here, to make sure
	 * it is secure and it is properly cleaned up.  Really just
	 * a temp file creation and clean up method wrapper for analyze_file()
	 *
	 * @param string $data the data to analyze
	 *
	 * @param string MIME type false for none.
	 *
	 * @author skwashd
	 */
	public static function analyze_data($data)
	{
		if(!is_writeable(@$GLOBALS['egw_info']['server']['temp_dir']))
		{
			//nothing we can do but bail out
			return false;
		}

		mt_srand(time());
		$filename = $GLOBALS['egw_info']['server']['temp_dir'] . '/'
			. md5( time() + mt_rand() ) . '.tmp';

		$fp = @fopen($filename, 'ab');
		if(!$fp || !$data)
		{
			//houston we have a problem - bail out
			return false;
		}

		if(!fwrite($fp, $data))
		{
			//bail out again
			return false;
		}
		fclose($fp);
		chmod($filename, 0600); //just to be cautious

		$mime = self::analyze_file($filename);

		unlink($filename);//remove the temp file

		return $mime;
	}

	/**
	 * Get an array containing a mapping of common file extensions to
	 * MIME types.
	 *
	 * Original array taken from http://horde.org
	 * added OpenOffice Extensions by KL@stylite.de
	 *
	 * @author skwashd
	 * @author ralfbecker - convert to a static array ;-)
	 *
	 * @var array of extention to mime mappings
	 */
	static $mime_extension_map = array(
		'txt'				=> 'text/plain',	// should be first, so text/plain maps back to .txt
		'3dml'              => 'text/vnd.in3d.3dml',
		'3ds'               => 'image/x-3ds',
		'3g2'               => 'video/3gpp2',
		'3ga'               => 'video/3gpp',
		'3gp'               => 'video/3gpp',
		'3gp2'              => 'video/3gpp2',
		'3gpp'              => 'video/3gpp',
		'3gpp2'             => 'video/3gpp2',
		'7z'                => 'application/x-7z-compressed',
		'C'                 => 'text/x-c++src',
		'Z'                 => 'application/x-compress',
		'a'                 => 'application/x-archive',
		'aab'               => 'application/x-authorware-bin',
		'aac'               => 'audio/x-aac',
		'aam'               => 'application/x-authorware-map',
		'aas'               => 'application/x-authorware-seg',
		'abw'               => 'application/x-abiword',
		'abw.crashed'       => 'application/x-abiword',
		'abw.gz'            => 'application/x-abiword',
		'ac'                => 'application/pkix-attr-cert',
		'ac3'               => 'audio/ac3',
		'acc'               => 'application/vnd.americandynamics.acc',
		'ace'               => 'application/x-ace-compressed',
		'acu'               => 'application/vnd.acucobol',
		'acutc'             => 'application/vnd.acucorp',
		'adb'               => 'text/x-adasrc',
		'adp'               => 'audio/adpcm',
		'ads'               => 'text/x-adasrc',
		'aep'               => 'application/vnd.audiograph',
		'afm'               => 'application/x-font-type1',
		'afp'               => 'application/vnd.ibm.modcap',
		'ag'                => 'image/x-applix-graphics',
		'ahead'             => 'application/vnd.ahead.space',
		'ai'                => 'application/postscript',
		'aif'               => 'audio/x-aiff',
		'aifc'              => 'audio/x-aiff',
		'aiff'              => 'audio/x-aiff',
		'aiffc'             => 'audio/x-aifc',
		'air'               => 'application/vnd.adobe.air-application-installer-package+zip',
		'ait'               => 'application/vnd.dvb.ait',
		'al'                => 'application/x-perl',
		'alz'               => 'application/x-alz',
		'ami'               => 'application/vnd.amiga.ami',
		'amr'               => 'audio/AMR',
		'amz'               => 'audio/x-amzxml',
		'ani'               => 'application/x-navi-animation',
		'anx'               => 'application/annodex',
		'ape'               => 'audio/x-ape',
		'apk'               => 'application/vnd.android.package-archive',
		'appcache'          => 'text/cache-manifest',
		'application'       => 'application/x-ms-application',
		'apr'               => 'application/vnd.lotus-approach',
		'ar'                => 'application/x-archive',
		'arc'               => 'application/x-freearc',
		'arj'               => 'application/x-arj',
		'arw'               => 'image/x-sony-arw',
		'as'                => 'application/x-applix-spreadsheet',
		'asc'               => 'application/pgp-signature',
		'asf'               => 'video/x-ms-asf',
		'asm'               => 'text/x-asm',
		'aso'               => 'application/vnd.accpac.simply.aso',
		'asp'               => 'application/x-asp',
		'ass'               => 'text/x-ssa',
		'asx'               => 'video/x-ms-asf',
		'atc'               => 'application/vnd.acucorp',
		'atom'              => 'application/atom+xml',
		'atomcat'           => 'application/atomcat+xml',
		'atomsvc'           => 'application/atomsvc+xml',
		'atx'               => 'application/vnd.antix.game-component',
		'au'                => 'audio/basic',
		'avf'               => 'video/x-msvideo',
		'avi'               => 'video/x-msvideo',
		'aw'                => 'application/applixware',
		'awb'               => 'audio/AMR-WB',
		'awk'               => 'application/x-awk',
		'axa'               => 'audio/annodex',
		'axv'               => 'video/annodex',
		'azf'               => 'application/vnd.airzip.filesecure.azf',
		'azs'               => 'application/vnd.airzip.filesecure.azs',
		'azw'               => 'application/vnd.amazon.ebook',
		'bak'               => 'application/x-trash',
		'bat'               => 'application/x-msdownload',
		'bcpio'             => 'application/x-bcpio',
		'bdf'               => 'application/x-font-bdf',
		'bdm'               => 'application/vnd.syncml.dm+wbxml',
		'bdmv'              => 'video/mp2t',
		'bed'               => 'application/vnd.realvnc.bed',
		'bh2'               => 'application/vnd.fujitsu.oasysprs',
		'bib'               => 'text/x-bibtex',
		'bin'               => 'application/octet-stream',
		'blb'               => 'application/x-blorb',
		'blend'             => 'application/x-blender',
		'blender'           => 'application/x-blender',
		'blorb'             => 'application/x-blorb',
		'bmi'               => 'application/vnd.bmi',
		'bmp'               => 'image/bmp',
		'book'              => 'application/vnd.framemaker',
		'box'               => 'application/vnd.previewsystems.box',
		'boz'               => 'application/x-bzip2',
		'bpk'               => 'application/octet-stream',
		'btif'              => 'image/prs.btif',
		'bz'                => 'application/x-bzip',
		'bz2'               => 'application/x-bzip2',
		'c'                 => 'text/x-c',
		'c++'               => 'text/x-c++src',
		'c11amc'            => 'application/vnd.cluetrust.cartomobile-config',
		'c11amz'            => 'application/vnd.cluetrust.cartomobile-config-pkg',
		'c4d'               => 'application/vnd.clonk.c4group',
		'c4f'               => 'application/vnd.clonk.c4group',
		'c4g'               => 'application/vnd.clonk.c4group',
		'c4p'               => 'application/vnd.clonk.c4group',
		'c4u'               => 'application/vnd.clonk.c4group',
		'cab'               => 'application/vnd.ms-cab-compressed',
		'caf'               => 'audio/x-caf',
		'cap'               => 'application/vnd.tcpdump.pcap',
		'car'               => 'application/vnd.curl.car',
		'cat'               => 'application/vnd.ms-pki.seccat',
		'cb7'               => 'application/x-cbr',
		'cba'               => 'application/x-cbr',
		'cbl'               => 'text/x-cobol',
		'cbr'               => 'application/x-cbr',
		'cbt'               => 'application/x-cbr',
		'cbz'               => 'application/x-cbr',
		'cc'                => 'text/x-c',
		'ccmx'              => 'application/x-ccmx',
		'cct'               => 'application/x-director',
		'ccxml'             => 'application/ccxml+xml',
		'cdbcmsg'           => 'application/vnd.contact.cmsg',
		'cdf'               => 'application/x-netcdf',
		'cdkey'             => 'application/vnd.mediastation.cdkey',
		'cdmia'             => 'application/cdmi-capability',
		'cdmic'             => 'application/cdmi-container',
		'cdmid'             => 'application/cdmi-domain',
		'cdmio'             => 'application/cdmi-object',
		'cdmiq'             => 'application/cdmi-queue',
		'cdr'               => 'application/vnd.corel-draw',
		'cdx'               => 'chemical/x-cdx',
		'cdxml'             => 'application/vnd.chemdraw+xml',
		'cdy'               => 'application/vnd.cinderella',
		'cer'               => 'application/pkix-cert',
		'cert'              => 'application/x-x509-ca-cert',
		'cfs'               => 'application/x-cfs-compressed',
		'cgm'               => 'image/cgm',
		'chat'              => 'application/x-chat',
		'chm'               => 'application/vnd.ms-htmlhelp',
		'chrt'              => 'application/vnd.kde.kchart',
		'cif'               => 'chemical/x-cif',
		'cii'               => 'application/vnd.anser-web-certificate-issue-initiation',
		'cil'               => 'application/vnd.ms-artgalry',
		'cla'               => 'application/vnd.claymore',
		'class'             => 'application/java-vm',
		'clkk'              => 'application/vnd.crick.clicker.keyboard',
		'clkp'              => 'application/vnd.crick.clicker.palette',
		'clkt'              => 'application/vnd.crick.clicker.template',
		'clkw'              => 'application/vnd.crick.clicker.wordbank',
		'clkx'              => 'application/vnd.crick.clicker',
		'clp'               => 'application/x-msclip',
		'clpi'              => 'video/mp2t',
		'cls'               => 'text/x-tex',
		'cmake'             => 'text/x-cmake',
		'cmc'               => 'application/vnd.cosmocaller',
		'cmdf'              => 'chemical/x-cmdf',
		'cml'               => 'chemical/x-cml',
		'cmp'               => 'application/vnd.yellowriver-custom-menu',
		'cmx'               => 'image/x-cmx',
		'cob'               => 'text/x-cobol',
		'cod'               => 'application/vnd.rim.cod',
		'com'               => 'application/x-msdownload',
		'conf'              => 'text/plain',
		'cpi'               => 'video/mp2t',
		'cpio'              => 'application/x-cpio',
		'cpio.gz'           => 'application/x-cpio-compressed',
		'cpp'               => 'text/x-c',
		'cpt'               => 'application/mac-compactpro',
		'cr2'               => 'image/x-canon-cr2',
		'crd'               => 'application/x-mscardfile',
		'crl'               => 'application/pkix-crl',
		'crt'               => 'application/x-x509-ca-cert',
		'crw'               => 'image/x-canon-crw',
		'cryptonote'        => 'application/vnd.rig.cryptonote',
		'cs'                => 'text/x-csharp',
		'csh'               => 'application/x-csh',
		'csml'              => 'chemical/x-csml',
		'csp'               => 'application/vnd.commonspace',
		'css'               => 'text/css',
		'cst'               => 'application/x-director',
		'csv'               => 'text/csv',
		'cu'                => 'application/cu-seeme',
		'cue'               => 'application/x-cue',
		'cur'               => 'image/x-win-bitmap',
		'curl'              => 'text/vnd.curl',
		'cww'               => 'application/prs.cww',
		'cxt'               => 'application/x-director',
		'cxx'               => 'text/x-c',
		'd'                 => 'text/x-dsrc',
		'dae'               => 'model/vnd.collada+xml',
		'daf'               => 'application/vnd.mobius.daf',
		'dar'               => 'application/x-dar',
		'dart'              => 'application/vnd.dart',
		'dataless'          => 'application/vnd.fdsn.seed',
		'davmount'          => 'application/davmount+xml',
		'dbf'               => 'application/x-dbf',
		'dbk'               => 'application/docbook+xml',
		'dc'                => 'application/x-dc-rom',
		'dcl'               => 'text/x-dcl',
		'dcm'               => 'application/dicom',
		'dcr'               => 'application/x-director',
		'dcurl'             => 'text/vnd.curl.dcurl',
		'dd2'               => 'application/vnd.oma.dd2+xml',
		'ddd'               => 'application/vnd.fujixerox.ddd',
		'dds'               => 'image/x-dds',
		'deb'               => 'application/x-debian-package',
		'def'               => 'text/plain',
		'deploy'            => 'application/octet-stream',
		'der'               => 'application/x-x509-ca-cert',
		'desktop'           => 'application/x-desktop',
		'dfac'              => 'application/vnd.dreamfactory',
		'dgc'               => 'application/x-dgc-compressed',
		'di'                => 'text/x-dsrc',
		'dia'               => 'application/x-dia-diagram',
		'dic'               => 'text/x-c',
		'diff'              => 'text/diff',
		'dir'               => 'application/x-director',
		'dis'               => 'application/vnd.mobius.dis',
		'dist'              => 'application/octet-stream',
		'distz'             => 'application/octet-stream',
		'divx'              => 'video/x-msvideo',
		'djv'               => 'image/vnd.djvu',
		'djvu'              => 'image/vnd.djvu',
		'dll'               => 'application/x-msdownload',
		'dmg'               => 'application/x-apple-diskimage',
		'dmp'               => 'application/vnd.tcpdump.pcap',
		'dms'               => 'application/octet-stream',
		'dna'               => 'application/vnd.dna',
		'dng'               => 'image/x-adobe-dng',
		'doc'               => 'application/msword',
		'docbook'           => 'application/x-docbook+xml',
		'docm'              => 'application/vnd.ms-word.document.macroenabled.12',
		'docx'              => 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
		'dot'               => 'application/msword',
		'dotm'              => 'application/vnd.ms-word.template.macroenabled.12',
		'dotx'              => 'application/vnd.openxmlformats-officedocument.wordprocessingml.template',
		'dp'                => 'application/vnd.osgi.dp',
		'dpg'               => 'application/vnd.dpgraph',
		'dra'               => 'audio/vnd.dra',
		'dsc'               => 'text/prs.lines.tag',
		'dsl'               => 'text/x-dsl',
		'dssc'              => 'application/dssc+der',
		'dtb'               => 'application/x-dtbook+xml',
		'dtd'               => 'application/xml-dtd',
		'dts'               => 'audio/vnd.dts',
		'dtshd'             => 'audio/vnd.dts.hd',
		'dtx'               => 'text/x-tex',
		'dump'              => 'application/octet-stream',
		'dv'                => 'video/dv',
		'dvb'               => 'video/vnd.dvb.file',
		'dvi'               => 'application/x-dvi',
		'dvi.bz2'           => 'application/x-bzdvi',
		'dvi.gz'            => 'application/x-gzdvi',
		'dwf'               => 'model/vnd.dwf',
		'dwg'               => 'image/vnd.dwg',
		'dxf'               => 'image/vnd.dxf',
		'dxp'               => 'application/vnd.spotfire.dxp',
		'dxr'               => 'application/x-director',
		'e'                 => 'text/x-eiffel',
		'ecelp4800'         => 'audio/vnd.nuera.ecelp4800',
		'ecelp7470'         => 'audio/vnd.nuera.ecelp7470',
		'ecelp9600'         => 'audio/vnd.nuera.ecelp9600',
		'ecma'              => 'application/ecmascript',
		'edm'               => 'application/vnd.novadigm.edm',
		'edx'               => 'application/vnd.novadigm.edx',
		'efif'              => 'application/vnd.picsel',
		'egon'              => 'application/x-egon',
		'ei6'               => 'application/vnd.pg.osasli',
		'eif'               => 'text/x-eiffel',
		'el'                => 'text/x-emacs-lisp',
		'elc'               => 'application/octet-stream',
		'emf'               => 'application/x-msmetafile',
		'eml'               => 'message/rfc822',
		'emma'              => 'application/emma+xml',
		'emp'               => 'application/vnd.emusic-emusic_package',
		'emz'               => 'application/x-msmetafile',
		'ent'               => 'application/xml-external-parsed-entity',
		'eol'               => 'audio/vnd.digital-winds',
		'eot'               => 'application/vnd.ms-fontobject',
		'eps'               => 'application/postscript',
		'eps.bz2'           => 'image/x-bzeps',
		'eps.gz'            => 'image/x-gzeps',
		'epsf'              => 'image/x-eps',
		'epsf.bz2'          => 'image/x-bzeps',
		'epsf.gz'           => 'image/x-gzeps',
		'epsi'              => 'image/x-eps',
		'epsi.bz2'          => 'image/x-bzeps',
		'epsi.gz'           => 'image/x-gzeps',
		'epub'              => 'application/epub+zip',
		'erl'               => 'text/x-erlang',
		'es'                => 'application/ecmascript',
		'es3'               => 'application/vnd.eszigno3+xml',
		'esa'               => 'application/vnd.osgi.subsystem',
		'esf'               => 'application/vnd.epson.esf',
		'et3'               => 'application/vnd.eszigno3+xml',
		'etheme'            => 'application/x-e-theme',
		'etx'               => 'text/x-setext',
		'eva'               => 'application/x-eva',
		'evy'               => 'application/x-envoy',
		'exe'               => 'application/x-msdownload',
		'exi'               => 'application/exi',
		'exr'               => 'image/x-exr',
		'ext'               => 'application/vnd.novadigm.ext',
		'ez'                => 'application/andrew-inset',
		'ez2'               => 'application/vnd.ezpix-album',
		'ez3'               => 'application/vnd.ezpix-package',
		'f'                 => 'text/x-fortran',
		'f4a'               => 'audio/mp4',
		'f4b'               => 'audio/x-m4b',
		'f4v'               => 'video/x-f4v',
		'f77'               => 'text/x-fortran',
		'f90'               => 'text/x-fortran',
		'f95'               => 'text/x-fortran',
		'fb2'               => 'application/x-fictionbook+xml',
		'fb2.zip'           => 'application/x-zip-compressed-fb2',
		'fbs'               => 'image/vnd.fastbidsheet',
		'fcdt'              => 'application/vnd.adobe.formscentral.fcdt',
		'fcs'               => 'application/vnd.isac.fcs',
		'fdf'               => 'application/vnd.fdf',
		'fe_launch'         => 'application/vnd.denovo.fcselayout-link',
		'fg5'               => 'application/vnd.fujitsu.oasysgp',
		'fgd'               => 'application/x-director',
		'fh'                => 'image/x-freehand',
		'fh4'               => 'image/x-freehand',
		'fh5'               => 'image/x-freehand',
		'fh7'               => 'image/x-freehand',
		'fhc'               => 'image/x-freehand',
		'fig'               => 'application/x-xfig',
		'fits'              => 'image/fits',
		'fl'                => 'application/x-fluid',
		'flac'              => 'audio/x-flac',
		'flc'               => 'video/x-flic',
		'fli'               => 'video/x-fli',
		'flo'               => 'application/vnd.micrografx.flo',
		'flv'               => 'video/x-flv',
		'flw'               => 'application/vnd.kde.kivio',
		'flx'               => 'text/vnd.fmi.flexstor',
		'fly'               => 'text/vnd.fly',
		'fm'                => 'application/vnd.framemaker',
		'fnc'               => 'application/vnd.frogans.fnc',
		'fo'                => 'text/x-xslfo',
		'fodg'              => 'application/vnd.oasis.opendocument.graphics-flat-xml',
		'fodp'              => 'application/vnd.oasis.opendocument.presentation-flat-xml',
		'fods'              => 'application/vnd.oasis.opendocument.spreadsheet-flat-xml',
		'fodt'              => 'application/vnd.oasis.opendocument.text-flat-xml',
		'for'               => 'text/x-fortran',
		'fpx'               => 'image/vnd.fpx',
		'frame'             => 'application/vnd.framemaker',
		'fsc'               => 'application/vnd.fsc.weblaunch',
		'fst'               => 'image/vnd.fst',
		'ftc'               => 'application/vnd.fluxtime.clip',
		'fti'               => 'application/vnd.anser-web-funds-transfer-initiation',
		'fvt'               => 'video/vnd.fvt',
		'fxm'               => 'video/x-javafx',
		'fxp'               => 'application/vnd.adobe.fxp',
		'fxpl'              => 'application/vnd.adobe.fxp',
		'fzs'               => 'application/vnd.fuzzysheet',
		'g2w'               => 'application/vnd.geoplan',
		'g3'                => 'image/g3fax',
		'g3w'               => 'application/vnd.geospace',
		'gac'               => 'application/vnd.groove-account',
		'gam'               => 'application/x-tads',
		'gb'                => 'application/x-gameboy-rom',
		'gba'               => 'application/x-gba-rom',
		'gbr'               => 'application/rpki-ghostbusters',
		'gca'               => 'application/x-gca-compressed',
		'gcrd'              => 'text/vcard',
		'gdl'               => 'model/vnd.gdl',
		'ged'               => 'application/x-gedcom',
		'gedcom'            => 'application/x-gedcom',
		'gem'               => 'application/x-tar',
		'gen'               => 'application/x-genesis-rom',
		'geo'               => 'application/vnd.dynageo',
		'gex'               => 'application/vnd.geometry-explorer',
		'gf'                => 'application/x-tex-gf',
		'gg'                => 'application/x-sms-rom',
		'ggb'               => 'application/vnd.geogebra.file',
		'ggt'               => 'application/vnd.geogebra.tool',
		'ghf'               => 'application/vnd.groove-help',
		'gif'               => 'image/gif',
		'gim'               => 'application/vnd.groove-identity-message',
		'glade'             => 'application/x-glade',
		'gml'               => 'application/gml+xml',
		'gmo'               => 'application/x-gettext-translation',
		'gmx'               => 'application/vnd.gmx',
		'gnc'               => 'application/x-gnucash',
		'gnd'               => 'application/gnunet-directory',
		'gnucash'           => 'application/x-gnucash',
		'gnumeric'          => 'application/x-gnumeric',
		'gnuplot'           => 'application/x-gnuplot',
		'go'                => 'text/x-go',
		'gp'                => 'application/x-gnuplot',
		'gpg'               => 'application/pgp-encrypted',
		'gph'               => 'application/vnd.flographit',
		'gplt'              => 'application/x-gnuplot',
		'gpx'               => 'application/gpx+xml',
		'gqf'               => 'application/vnd.grafeq',
		'gqs'               => 'application/vnd.grafeq',
		'gra'               => 'application/x-graphite',
		'gram'              => 'application/srgs',
		'gramps'            => 'application/x-gramps-xml',
		'gre'               => 'application/vnd.geometry-explorer',
		'grv'               => 'application/vnd.groove-injector',
		'grxml'             => 'application/srgs+xml',
		'gsf'               => 'application/x-font-ghostscript',
		'gsm'               => 'audio/x-gsm',
		'gtar'              => 'application/x-gtar',
		'gtm'               => 'application/vnd.groove-tool-message',
		'gtw'               => 'model/vnd.gtw',
		'gv'                => 'text/vnd.graphviz',
		'gvp'               => 'text/x-google-video-pointer',
		'gxf'               => 'application/gxf',
		'gxt'               => 'application/vnd.geonext',
		'gz'                => 'application/x-gzip',
		'h'                 => 'text/x-c',
		'h++'               => 'text/x-c++hdr',
		'h261'              => 'video/h261',
		'h263'              => 'video/h263',
		'h264'              => 'video/h264',
		'h4'                => 'application/x-hdf',
		'h5'                => 'application/x-hdf',
		'hal'               => 'application/vnd.hal+xml',
		'hbci'              => 'application/vnd.hbci',
		'hdf'               => 'application/x-hdf',
		'hdf4'              => 'application/x-hdf',
		'hdf5'              => 'application/x-hdf',
		'hh'                => 'text/x-c',
		'hlp'               => 'application/winhlp',
		'hp'                => 'text/x-c++hdr',
		'hpgl'              => 'application/vnd.hp-hpgl',
		'hpid'              => 'application/vnd.hp-hpid',
		'hpp'               => 'text/x-c++hdr',
		'hps'               => 'application/vnd.hp-hps',
		'hqx'               => 'application/mac-binhex40',
		'hs'                => 'text/x-haskell',
		'htke'              => 'application/vnd.kenameaapp',
		'htm'               => 'text/html',
		'html'              => 'text/html',
		'hvd'               => 'application/vnd.yamaha.hv-dic',
		'hvp'               => 'application/vnd.yamaha.hv-voice',
		'hvs'               => 'application/vnd.yamaha.hv-script',
		'hwp'               => 'application/x-hwp',
		'hwt'               => 'application/x-hwt',
		'hxx'               => 'text/x-c++hdr',
		'i2g'               => 'application/vnd.intergeo',
		'ica'               => 'application/x-ica',
		'icb'               => 'image/x-tga',
		'icc'               => 'application/vnd.iccprofile',
		'ice'               => 'x-conference/x-cooltalk',
		'icm'               => 'application/vnd.iccprofile',
		'icns'              => 'image/x-icns',
		'ico'               => 'image/x-icon',
		'ics'               => 'text/calendar',
		'idl'               => 'text/x-idl',
		'ief'               => 'image/ief',
		'ifb'               => 'text/calendar',
		'iff'               => 'image/x-ilbm',
		'ifm'               => 'application/vnd.shana.informed.formdata',
		'iges'              => 'model/iges',
		'igl'               => 'application/vnd.igloader',
		'igm'               => 'application/vnd.insors.igm',
		'igs'               => 'model/iges',
		'igx'               => 'application/vnd.micrografx.igx',
		'iif'               => 'application/vnd.shana.informed.interchange',
		'ilbm'              => 'image/x-ilbm',
		'ime'               => 'text/x-iMelody',
		'img'               => 'application/x-raw-disk-image',
		'img.xz'            => 'application/x-raw-disk-image-xz-compressed',
		'imp'               => 'application/vnd.accpac.simply.imp',
		'ims'               => 'application/vnd.ms-ims',
		'imy'               => 'text/x-iMelody',
		'in'                => 'text/plain',
		'ink'               => 'application/inkml+xml',
		'inkml'             => 'application/inkml+xml',
		'ins'               => 'text/x-tex',
		'install'           => 'application/x-install-instructions',
		'iota'              => 'application/vnd.astraea-software.iota',
		'ipfix'             => 'application/ipfix',
		'ipk'               => 'application/vnd.shana.informed.package',
		'iptables'          => 'text/x-iptables',
		'irm'               => 'application/vnd.ibm.rights-management',
		'irp'               => 'application/vnd.irepository.package+xml',
		'iso'               => 'application/x-iso9660-image',
		'iso9660'           => 'application/x-cd-image',
		'it'                => 'audio/x-it',
		'it87'              => 'application/x-it87',
		'itp'               => 'application/vnd.shana.informed.formtemplate',
		'ivp'               => 'application/vnd.immervision-ivp',
		'ivu'               => 'application/vnd.immervision-ivu',
		'jad'               => 'text/vnd.sun.j2me.app-descriptor',
		'jam'               => 'application/vnd.jam',
		'jar'               => 'application/java-archive',
		'java'              => 'text/x-java-source',
		'jceks'             => 'application/x-java-jce-keystore',
		'jisp'              => 'application/vnd.jisp',
		'jks'               => 'application/x-java-keystore',
		'jlt'               => 'application/vnd.hp-jlyt',
		'jng'               => 'image/x-jng',
		'jnlp'              => 'application/x-java-jnlp-file',
		'joda'              => 'application/vnd.joost.joda-archive',
		'jp2'               => 'image/jp2',
		'jpg'               => 'image/jpeg',
		'jpeg'              => 'image/jpeg',
		'jpf'               => 'image/jp2',
		'jpe'               => 'image/jpeg',
		'jpgm'              => 'video/jpm',
		'jpgv'              => 'video/jpeg',
		'jpm'               => 'video/jpm',
		'jpr'               => 'application/x-jbuilder-project',
		'jpx'               => 'application/x-jbuilder-project',
		'js'                => 'application/javascript',
		'jsm'               => 'application/javascript',
		'json'              => 'application/json',
		'jsonml'            => 'application/jsonml+json',
		'k25'               => 'image/x-kodak-k25',
		'kar'               => 'audio/midi',
		'karbon'            => 'application/vnd.kde.karbon',
		'kdc'               => 'image/x-kodak-kdc',
		'kdelnk'            => 'application/x-desktop',
		'kexi'              => 'application/x-kexiproject-sqlite2',
		'kexic'             => 'application/x-kexi-connectiondata',
		'kexis'             => 'application/x-kexiproject-shortcut',
		'key'               => 'application/x-iwork-keynote-sffkey',
		'kfo'               => 'application/vnd.kde.kformula',
		'kia'               => 'application/vnd.kidspiration',
		'kil'               => 'application/x-killustrator',
		'kino'              => 'application/smil',
		'kml'               => 'application/vnd.google-earth.kml+xml',
		'kmz'               => 'application/vnd.google-earth.kmz',
		'kne'               => 'application/vnd.kinar',
		'knp'               => 'application/vnd.kinar',
		'kon'               => 'application/vnd.kde.kontour',
		'kpm'               => 'application/x-kpovmodeler',
		'kpr'               => 'application/vnd.kde.kpresenter',
		'kpt'               => 'application/vnd.kde.kpresenter',
		'kpxx'              => 'application/vnd.ds-keypoint',
		'kra'               => 'application/x-krita',
		'ks'                => 'application/x-java-keystore',
		'ksp'               => 'application/vnd.kde.kspread',
		'ktr'               => 'application/vnd.kahootz',
		'ktx'               => 'image/ktx',
		'ktz'               => 'application/vnd.kahootz',
		'kud'               => 'application/x-kugar',
		'kwd'               => 'application/vnd.kde.kword',
		'kwt'               => 'application/vnd.kde.kword',
		'la'                => 'application/x-shared-library-la',
		'lasxml'            => 'application/vnd.las.las+xml',
		'latex'             => 'application/x-latex',
		'lbd'               => 'application/vnd.llamagraphics.life-balance.desktop',
		'lbe'               => 'application/vnd.llamagraphics.life-balance.exchange+xml',
		'lbm'               => 'image/x-ilbm',
		'ldif'              => 'text/x-ldif',
		'les'               => 'application/vnd.hhe.lesson-player',
		'lha'               => 'application/x-lzh-compressed',
		'lhs'               => 'text/x-literate-haskell',
		'lhz'               => 'application/x-lhz',
		'link66'            => 'application/vnd.route66.link66+xml',
		'list'              => 'text/plain',
		'list3820'          => 'application/vnd.ibm.modcap',
		'listafp'           => 'application/vnd.ibm.modcap',
		'lnk'               => 'application/x-ms-shortcut',
		'log'               => 'text/plain',
		'lostxml'           => 'application/lost+xml',
		'lrf'               => 'application/octet-stream',
		'lrm'               => 'application/vnd.ms-lrm',
		'lrz'               => 'application/x-lrzip',
		'ltf'               => 'application/vnd.frogans.ltf',
		'ltx'               => 'text/x-tex',
		'lua'               => 'text/x-lua',
		'lvp'               => 'audio/vnd.lucent.voice',
		'lwo'               => 'image/x-lwo',
		'lwob'              => 'image/x-lwo',
		'lwp'               => 'application/vnd.lotus-wordpro',
		'lws'               => 'image/x-lws',
		'ly'                => 'text/x-lilypond',
		'lyx'               => 'application/x-lyx',
		'lz'                => 'application/x-lzip',
		'lz4'               => 'application/x-lz4',
		'lzh'               => 'application/x-lzh-compressed',
		'lzma'              => 'application/x-lzma',
		'lzo'               => 'application/x-lzop',
		'm'                 => 'text/x-objcsrc',
		'm13'               => 'application/x-msmediaview',
		'm14'               => 'application/x-msmediaview',
		'm15'               => 'audio/x-mod',
		'm1u'               => 'video/vnd.mpegurl',
		'm1v'               => 'video/mpeg',
		'm21'               => 'application/mp21',
		'm2a'               => 'audio/mpeg',
		'm2t'               => 'video/mp2t',
		'm2ts'              => 'video/mp2t',
		'm2v'               => 'video/mpeg',
		'm3a'               => 'audio/mpeg',
		'm3u'               => 'audio/x-mpegurl',
		'm3u8'              => 'application/vnd.apple.mpegurl',
		'm4'                => 'application/x-m4',
		'm4a'               => 'audio/mp4',
		'm4b'               => 'audio/x-m4b',
		'm4u'               => 'video/vnd.mpegurl',
		'm4v'               => 'video/x-m4v',
		'ma'                => 'application/mathematica',
		'mab'               => 'application/x-markaby',
		'mads'              => 'application/mads+xml',
		'mag'               => 'application/vnd.ecowin.chart',
		'mak'               => 'text/x-makefile',
		'maker'             => 'application/vnd.framemaker',
		'man'               => 'text/troff',
		'manifest'          => 'text/cache-manifest',
		'mar'               => 'application/octet-stream',
		'markdown'          => 'text/x-markdown',
		'mathml'            => 'application/mathml+xml',
		'mb'                => 'application/mathematica',
		'mbk'               => 'application/vnd.mobius.mbk',
		'mbox'              => 'application/mbox',
		'mc1'               => 'application/vnd.medcalcdata',
		'mcd'               => 'application/vnd.mcd',
		'mcurl'             => 'text/vnd.curl.mcurl',
		'md'                => 'text/x-markdown',
		'mdb'               => 'application/x-msaccess',
		'mdi'               => 'image/vnd.ms-modi',
		'me'                => 'text/troff',
		'med'               => 'audio/x-mod',
		'mesh'              => 'model/mesh',
		'meta4'             => 'application/metalink4+xml',
		'metalink'          => 'application/metalink+xml',
		'mets'              => 'application/mets+xml',
		'mfm'               => 'application/vnd.mfmp',
		'mft'               => 'application/rpki-manifest',
		'mgp'               => 'application/vnd.osgeo.mapguide.package',
		'mgz'               => 'application/vnd.proteus.magazine',
		'mht'               => 'application/x-mimearchive',
		'mhtml'             => 'application/x-mimearchive',
		'mid'               => 'audio/midi',
		'midi'              => 'audio/midi',
		'mie'               => 'application/x-mie',
		'mif'               => 'application/vnd.mif',
		'mime'              => 'message/rfc822',
		'minipsf'           => 'audio/x-minipsf',
		'mj2'               => 'video/mj2',
		'mjp2'              => 'video/mj2',
		'mk'                => 'text/x-makefile',
		'mk3d'              => 'video/x-matroska',
		'mka'               => 'audio/x-matroska',
		'mkd'               => 'text/x-markdown',
		'mks'               => 'video/x-matroska',
		'mkv'               => 'video/x-matroska',
		'ml'                => 'text/x-ocaml',
		'mli'               => 'text/x-ocaml',
		'mlp'               => 'application/vnd.dolby.mlp',
		'mm'                => 'text/x-troff-mm',
		'mmd'               => 'application/vnd.chipnuts.karaoke-mmd',
		'mmf'               => 'application/vnd.smaf',
		'mml'               => 'application/mathml+xml',
		'mmr'               => 'image/vnd.fujixerox.edmics-mmr',
		'mng'               => 'video/x-mng',
		'mny'               => 'application/x-msmoney',
		'mo'                => 'application/x-gettext-translation',
		'mo3'               => 'audio/x-mo3',
		'mobi'              => 'application/x-mobipocket-ebook',
		'moc'               => 'text/x-moc',
		'mod'               => 'audio/x-mod',
		'mods'              => 'application/mods+xml',
		'mof'               => 'text/x-mof',
		'moov'              => 'video/quicktime',
		'mov'               => 'video/quicktime',
		'movie'             => 'video/x-sgi-movie',
		'mp+'               => 'audio/x-musepack',
		'mp2'               => 'audio/mpeg',
		'mp21'              => 'application/mp21',
		'mp2a'              => 'audio/mpeg',
		'mp3'               => 'audio/mpeg',
		'mp4'               => 'video/mp4',
		'mp4a'              => 'audio/mp4',
		'mp4s'              => 'application/mp4',
		'mp4v'              => 'video/mp4',
		'mpc'               => 'application/vnd.mophun.certificate',
		'mpe'               => 'video/mpeg',
		'mpeg'              => 'video/mpeg',
		'mpg'               => 'video/mpeg',
		'mpg4'              => 'video/mp4',
		'mpga'              => 'audio/mpeg',
		'mpkg'              => 'application/vnd.apple.installer+xml',
		'mpl'               => 'video/mp2t',
		'mpls'              => 'video/mp2t',
		'mpm'               => 'application/vnd.blueice.multipass',
		'mpn'               => 'application/vnd.mophun.application',
		'mpp'               => 'application/vnd.ms-project',
		'mpt'               => 'application/vnd.ms-project',
		'mpy'               => 'application/vnd.ibm.minipay',
		'mqy'               => 'application/vnd.mobius.mqy',
		'mrc'               => 'application/marc',
		'mrcx'              => 'application/marcxml+xml',
		'mrl'               => 'text/x-mrml',
		'mrml'              => 'text/x-mrml',
		'mrw'               => 'image/x-minolta-mrw',
		'ms'                => 'text/troff',
		'mscml'             => 'application/mediaservercontrol+xml',
		'mseed'             => 'application/vnd.fdsn.mseed',
		'mseq'              => 'application/vnd.mseq',
		'msf'               => 'application/vnd.epson.msf',
		'msh'               => 'model/mesh',
		'msi'               => 'application/x-msdownload',
		'msl'               => 'application/vnd.mobius.msl',
		'msod'              => 'image/x-msod',
		'msty'              => 'application/vnd.muvee.style',
		'msx'               => 'application/x-msx-rom',
		'mtm'               => 'audio/x-mod',
		'mts'               => 'model/vnd.mts',
		'mup'               => 'text/x-mup',
		'mus'               => 'application/vnd.musician',
		'musicxml'          => 'application/vnd.recordare.musicxml+xml',
		'mvb'               => 'application/x-msmediaview',
		'mwf'               => 'application/vnd.mfer',
		'mxf'               => 'application/mxf',
		'mxl'               => 'application/vnd.recordare.musicxml',
		'mxml'              => 'application/xv+xml',
		'mxs'               => 'application/vnd.triscape.mxs',
		'mxu'               => 'video/vnd.mpegurl',
		'n-gage'            => 'application/vnd.nokia.n-gage.symbian.install',
		'n3'                => 'text/n3',
		'n64'               => 'application/x-n64-rom',
		'nb'                => 'application/mathematica',
		'nbp'               => 'application/vnd.wolfram.player',
		'nc'                => 'application/x-netcdf',
		'ncx'               => 'application/x-dtbncx+xml',
		'nds'               => 'application/x-nintendo-ds-rom',
		'nef'               => 'image/x-nikon-nef',
		'nes'               => 'application/x-nes-rom',
		'nfo'               => 'text/x-nfo',
		'ngdat'             => 'application/vnd.nokia.n-gage.data',
		'nitf'              => 'application/vnd.nitf',
		'nlu'               => 'application/vnd.neurolanguage.nlu',
		'nml'               => 'application/vnd.enliven',
		'nnd'               => 'application/vnd.noblenet-directory',
		'nns'               => 'application/vnd.noblenet-sealer',
		'nnw'               => 'application/vnd.noblenet-web',
		'not'               => 'text/x-mup',
		'npx'               => 'image/vnd.net-fpx',
		'nsc'               => 'application/x-conference',
		'nsf'               => 'application/vnd.lotus-notes',
		'nsv'               => 'video/x-nsv',
		'ntf'               => 'application/vnd.nitf',
		'nzb'               => 'application/x-nzb',
		'o'                 => 'application/x-object',
		'oa2'               => 'application/vnd.fujitsu.oasys2',
		'oa3'               => 'application/vnd.fujitsu.oasys3',
		'oas'               => 'application/vnd.fujitsu.oasys',
		'obd'               => 'application/x-msbinder',
		'obj'               => 'application/x-tgif',
		'ocl'               => 'text/x-ocl',
		'oda'               => 'application/oda',
		'odb'               => 'application/vnd.oasis.opendocument.database',
		'odc'               => 'application/vnd.oasis.opendocument.chart',
		'odf'               => 'application/vnd.oasis.opendocument.formula',
		'odft'              => 'application/vnd.oasis.opendocument.formula-template',
		'odg'               => 'application/vnd.oasis.opendocument.graphics',
		'odi'               => 'application/vnd.oasis.opendocument.image',
		'odm'               => 'application/vnd.oasis.opendocument.text-master',
		'odp'               => 'application/vnd.oasis.opendocument.presentation',
		'ods'               => 'application/vnd.oasis.opendocument.spreadsheet',
		'odt'               => 'application/vnd.oasis.opendocument.text',
		'oga'               => 'audio/ogg',
		'ogg'               => 'audio/ogg',
		'ogm'               => 'video/x-ogm+ogg',
		'ogv'               => 'video/ogg',
		'ogx'               => 'application/ogg',
		'old'               => 'application/x-trash',
		'oleo'              => 'application/x-oleo',
		'omdoc'             => 'application/omdoc+xml',
		'onepkg'            => 'application/onenote',
		'onetmp'            => 'application/onenote',
		'onetoc'            => 'application/onenote',
		'onetoc2'           => 'application/onenote',
		'ooc'               => 'text/x-ooc',
		'oot'               => 'application/vnd.oasis.opendocument.text',
		'opf'               => 'application/oebps-package+xml',
		'opml'              => 'text/x-opml',
		'oprc'              => 'application/vnd.palm',
		'opus'              => 'audio/ogg',
		'ora'               => 'image/openraster',
		'orf'               => 'image/x-olympus-orf',
		'org'               => 'application/vnd.lotus-organizer',
		'osf'               => 'application/vnd.yamaha.openscoreformat',
		'osfpvg'            => 'application/vnd.yamaha.openscoreformat.osfpvg+xml',
		'otc'               => 'application/vnd.oasis.opendocument.chart-template',
		'otf'               => 'application/x-font-otf',
		'otg'               => 'application/vnd.oasis.opendocument.graphics-template',
		'oth'               => 'application/vnd.oasis.opendocument.text-web',
		'oti'               => 'application/vnd.oasis.opendocument.image-template',
		'otp'               => 'application/vnd.oasis.opendocument.presentation-template',
		'ots'               => 'application/vnd.oasis.opendocument.spreadsheet-template',
		'ott'               => 'application/vnd.oasis.opendocument.text-template',
		'owl'               => 'application/rdf+xml',
		'oxps'              => 'application/oxps',
		'oxt'               => 'application/vnd.openofficeorg.extension',
		'p'                 => 'text/x-pascal',
		'p10'               => 'application/pkcs10',
		'p12'               => 'application/x-pkcs12',
		'p7b'               => 'application/x-pkcs7-certificates',
		'p7c'               => 'application/pkcs7-mime',
		'p7m'               => 'application/pkcs7-mime',
		'p7r'               => 'application/x-pkcs7-certreqresp',
		'p7s'               => 'application/pkcs7-signature',
		'p8'                => 'application/pkcs8',
		'pack'              => 'application/x-java-pack200',
		'pak'               => 'application/x-pak',
		'par2'              => 'application/x-par2',
		'pas'               => 'text/x-pascal',
		'patch'             => 'text/diff',
		'paw'               => 'application/vnd.pawaafile',
		'pbd'               => 'application/vnd.powerbuilder6',
		'pbm'               => 'image/x-portable-bitmap',
		'pcap'              => 'application/vnd.tcpdump.pcap',
		'pcd'               => 'image/x-photo-cd',
		'pce'               => 'application/x-pc-engine-rom',
		'pcf'               => 'application/x-font-pcf',
		'pcf.gz'            => 'application/x-font-pcf',
		'pcf.z'             => 'application/x-font-pcf',
		'pcl'               => 'application/vnd.hp-pcl',
		'pclxl'             => 'application/vnd.hp-pclxl',
		'pct'               => 'image/x-pict',
		'pcurl'             => 'application/vnd.curl.pcurl',
		'pcx'               => 'image/x-pcx',
		'pdb'               => 'application/vnd.palm',
		'pdc'               => 'application/x-aportisdoc',
		'pdf'               => 'application/pdf',
		'pdf.bz2'           => 'application/x-bzpdf',
		'pdf.gz'            => 'application/x-gzpdf',
		'pdf.xz'            => 'application/x-xzpdf',
		'pef'               => 'image/x-pentax-pef',
		'pem'               => 'application/x-x509-ca-cert',
		'perl'              => 'application/x-perl',
		'pfa'               => 'application/x-font-type1',
		'pfb'               => 'application/x-font-type1',
		'pfm'               => 'application/x-font-type1',
		'pfr'               => 'application/font-tdpfr',
		'pfx'               => 'application/x-pkcs12',
		'pgm'               => 'image/x-portable-graymap',
		'pgn'               => 'application/x-chess-pgn',
		'pgp'               => 'application/pgp-encrypted',
		'php'               => 'application/x-httpd-php',
		'php3'              => 'application/x-httpd-php3',
		'php4'              => 'application/x-php',
		'php5'              => 'application/x-php',
		'phps'              => 'application/x-php',
		'pic'               => 'image/x-pict',
		'pict'              => 'image/x-pict',
		'pict1'             => 'image/x-pict',
		'pict2'             => 'image/x-pict',
		'pk'                => 'application/x-tex-pk',
		'pkg'               => 'application/octet-stream',
		'pki'               => 'application/pkixcmp',
		'pkipath'           => 'application/pkix-pkipath',
		'pkr'               => 'application/pgp-keys',
		'pl'                => 'application/x-perl',
		'pla'               => 'audio/x-iriver-pla',
		'plb'               => 'application/vnd.3gpp.pic-bw-large',
		'plc'               => 'application/vnd.mobius.plc',
		'plf'               => 'application/vnd.pocketlearn',
		'pln'               => 'application/x-planperfect',
		'pls'               => 'application/pls+xml',
		'pm'                => 'application/x-perl',
		'pml'               => 'application/vnd.ctc-posml',
		'png'               => 'image/png',
		'pnm'               => 'image/x-portable-anymap',
		'pntg'              => 'image/x-macpaint',
		'po'                => 'text/plain',
		'pod'               => 'application/x-perl',
		'por'               => 'application/x-spss-por',
		'portpkg'           => 'application/vnd.macports.portpkg',
		'pot'               => 'application/vnd.ms-powerpoint',
		'potm'              => 'application/vnd.ms-powerpoint.template.macroenabled.12',
		'potx'              => 'application/vnd.openxmlformats-officedocument.presentationml.template',
		'ppam'              => 'application/vnd.ms-powerpoint.addin.macroenabled.12',
		'ppd'               => 'application/vnd.cups-ppd',
		'ppm'               => 'image/x-portable-pixmap',
		'pps'               => 'application/vnd.ms-powerpoint',
		'ppsm'              => 'application/vnd.ms-powerpoint.slideshow.macroenabled.12',
		'ppsx'              => 'application/vnd.openxmlformats-officedocument.presentationml.slideshow',
		'ppt'               => 'application/vnd.ms-powerpoint',
		'pptm'              => 'application/vnd.ms-powerpoint.presentation.macroenabled.12',
		'pptx'              => 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
		'ppz'               => 'application/vnd.ms-powerpoint',
		'pqa'               => 'application/vnd.palm',
		'prc'               => 'application/x-mobipocket-ebook',
		'pre'               => 'application/vnd.lotus-freelance',
		'prf'               => 'application/pics-rules',
		'ps'                => 'application/postscript',
		'ps.bz2'            => 'application/x-bzpostscript',
		'ps.gz'             => 'application/x-gzpostscript',
		'psb'               => 'application/vnd.3gpp.pic-bw-small',
		'psd'               => 'image/vnd.adobe.photoshop',
		'psf'               => 'application/x-font-linux-psf',
		'psf.gz'            => 'application/x-gz-font-linux-psf',
		'psflib'            => 'audio/x-psflib',
		'psid'              => 'audio/prs.sid',
		'pskcxml'           => 'application/pskc+xml',
		'psw'               => 'application/x-pocket-word',
		'ptid'              => 'application/vnd.pvi.ptid1',
		'pub'               => 'application/x-mspublisher',
		'pvb'               => 'application/vnd.3gpp.pic-bw-var',
		'pw'                => 'application/x-pw',
		'pwn'               => 'application/vnd.3m.post-it-notes',
		'py'                => 'text/x-python',
		'pya'               => 'audio/vnd.ms-playready.media.pya',
		'pyc'               => 'application/x-python-bytecode',
		'pyo'               => 'application/x-python-bytecode',
		'pyv'               => 'video/vnd.ms-playready.media.pyv',
		'pyx'               => 'text/x-python',
		'qam'               => 'application/vnd.epson.quickanime',
		'qbo'               => 'application/vnd.intu.qbo',
		'qfx'               => 'application/vnd.intu.qfx',
		'qif'               => 'application/x-qw',
		'qml'               => 'text/x-qml',
		'qps'               => 'application/vnd.publishare-delta-tree',
		'qt'                => 'video/quicktime',
		'qti'               => 'application/x-qtiplot',
		'qti.gz'            => 'application/x-qtiplot',
		'qtif'              => 'image/x-quicktime',
		'qtl'               => 'application/x-quicktime-media-link',
		'qtvr'              => 'video/quicktime',
		'qwd'               => 'application/vnd.quark.quarkxpress',
		'qwt'               => 'application/vnd.quark.quarkxpress',
		'qxb'               => 'application/vnd.quark.quarkxpress',
		'qxd'               => 'application/vnd.quark.quarkxpress',
		'qxl'               => 'application/vnd.quark.quarkxpress',
		'qxt'               => 'application/vnd.quark.quarkxpress',
		'ra'                => 'audio/x-pn-realaudio',
		'raf'               => 'image/x-fuji-raf',
		'ram'               => 'audio/x-pn-realaudio',
		'rar'               => 'application/x-rar-compressed',
		'ras'               => 'image/x-cmu-raster',
		'raw'               => 'image/x-panasonic-raw',
		'raw-disk-image'    => 'application/x-raw-disk-image',
		'raw-disk-image.xz' => 'application/x-raw-disk-image-xz-compressed',
		'rax'               => 'audio/vnd.rn-realaudio',
		'rb'                => 'application/x-ruby',
		'rcprofile'         => 'application/vnd.ipunplugged.rcprofile',
		'rdf'               => 'application/rdf+xml',
		'rdfs'              => 'application/rdf+xml',
		'rdz'               => 'application/vnd.data-vision.rdz',
		'reg'               => 'text/x-ms-regedit',
		'rej'               => 'text/x-reject',
		'rep'               => 'application/vnd.businessobjects',
		'res'               => 'application/x-dtbresource+xml',
		'rgb'               => 'image/x-rgb',
		'rif'               => 'application/reginfo+xml',
		'rip'               => 'audio/vnd.rip',
		'ris'               => 'application/x-research-info-systems',
		'rl'                => 'application/resource-lists+xml',
		'rlc'               => 'image/vnd.fujixerox.edmics-rlc',
		'rld'               => 'application/resource-lists-diff+xml',
		'rle'               => 'image/rle',
		'rm'                => 'application/vnd.rn-realmedia',
		'rmi'               => 'audio/midi',
		'rmj'               => 'application/vnd.rn-realmedia',
		'rmm'               => 'application/vnd.rn-realmedia',
		'rmp'               => 'audio/x-pn-realaudio-plugin',
		'rms'               => 'application/vnd.jcp.javame.midlet-rms',
		'rmvb'              => 'application/vnd.rn-realmedia-vbr',
		'rmx'               => 'application/vnd.rn-realmedia',
		'rnc'               => 'application/relax-ng-compact-syntax',
		'rng'               => 'application/xml',
		'roa'               => 'application/rpki-roa',
		'roff'              => 'text/troff',
		'rp'                => 'image/vnd.rn-realpix',
		'rp9'               => 'application/vnd.cloanto.rp9',
		'rpm'               => 'application/x-rpm',
		'rpss'              => 'application/vnd.nokia.radio-presets',
		'rpst'              => 'application/vnd.nokia.radio-preset',
		'rq'                => 'application/sparql-query',
		'rs'                => 'application/rls-services+xml',
		'rsd'               => 'application/rsd+xml',
		'rss'               => 'application/rss+xml',
		'rt'                => 'text/vnd.rn-realtext',
		'rtf'               => 'application/rtf',
		'rtx'               => 'text/richtext',
		'rv'                => 'video/vnd.rn-realvideo',
		'rvx'               => 'video/vnd.rn-realvideo',
		'rw2'               => 'image/x-panasonic-raw2',
		's'                 => 'text/x-asm',
		's3m'               => 'audio/s3m',
		'saf'               => 'application/vnd.yamaha.smaf-audio',
		'sam'               => 'application/x-amipro',
		'sami'              => 'application/x-sami',
		'sav'               => 'application/x-spss-sav',
		'sbml'              => 'application/sbml+xml',
		'sc'                => 'application/vnd.ibm.secure-container',
		'scala'             => 'text/x-scala',
		'scd'               => 'application/x-msschedule',
		'scm'               => 'application/vnd.lotus-screencam',
		'scq'               => 'application/scvp-cv-request',
		'scs'               => 'application/scvp-cv-response',
		'scurl'             => 'text/vnd.curl.scurl',
		'sda'               => 'application/vnd.stardivision.draw',
		'sdc'               => 'application/vnd.stardivision.calc',
		'sdd'               => 'application/vnd.stardivision.impress',
		'sdkd'              => 'application/vnd.solent.sdkm+xml',
		'sdkm'              => 'application/vnd.solent.sdkm+xml',
		'sdp'               => 'application/sdp',
		'sds'               => 'application/vnd.stardivision.chart',
		'sdw'               => 'application/vnd.stardivision.writer',
		'see'               => 'application/vnd.seemail',
		'seed'              => 'application/vnd.fdsn.seed',
		'sema'              => 'application/vnd.sema',
		'semd'              => 'application/vnd.semd',
		'semf'              => 'application/vnd.semf',
		'ser'               => 'application/java-serialized-object',
		'setpay'            => 'application/set-payment-initiation',
		'setreg'            => 'application/set-registration-initiation',
		'sfc'               => 'application/vnd.nintendo.snes.rom',
		'sfd-hdstx'         => 'application/vnd.hydrostatix.sof-data',
		'sfs'               => 'application/vnd.spotfire.sfs',
		'sfv'               => 'text/x-sfv',
		'sgf'               => 'application/x-go-sgf',
		'sgi'               => 'image/sgi',
		'sgl'               => 'application/vnd.stardivision.writer-global',
		'sgm'               => 'text/sgml',
		'sgml'              => 'text/sgml',
		'sh'                => 'application/x-sh',
		'shape'             => 'application/x-dia-shape',
		'shar'              => 'application/x-shar',
		'shf'               => 'application/shf+xml',
		'shn'               => 'application/x-shorten',
		'shtml'             => 'text/html',
		'siag'              => 'application/x-siag',
		'sid'               => 'image/x-mrsid-image',
		'sig'               => 'application/pgp-signature',
		'sik'               => 'application/x-trash',
		'sil'               => 'audio/silk',
		'silo'              => 'model/mesh',
		'sis'               => 'application/vnd.symbian.install',
		'sisx'              => 'application/vnd.symbian.install',
		'sit'               => 'application/x-stuffit',
		'sitx'              => 'application/x-stuffitx',
		'siv'               => 'application/sieve',
		'sk'                => 'image/x-skencil',
		'sk1'               => 'image/x-skencil',
		'skd'               => 'application/vnd.koan',
		'skm'               => 'application/vnd.koan',
		'skp'               => 'application/vnd.koan',
		'skr'               => 'application/pgp-keys',
		'skt'               => 'application/vnd.koan',
		'sldm'              => 'application/vnd.ms-powerpoint.slide.macroenabled.12',
		'sldx'              => 'application/vnd.openxmlformats-officedocument.presentationml.slide',
		'slk'               => 'text/spreadsheet',
		'slt'               => 'application/vnd.epson.salt',
		'sm'                => 'application/vnd.stepmania.stepchart',
		'smaf'              => 'application/x-smaf',
		'smc'               => 'application/vnd.nintendo.snes.rom',
		'smd'               => 'application/vnd.stardivision.mail',
		'smf'               => 'application/vnd.stardivision.math',
		'smi'               => 'application/smil+xml',
		'smil'              => 'application/smil+xml',
		'sml'               => 'application/smil',
		'sms'               => 'application/x-sms-rom',
		'smv'               => 'video/x-smv',
		'smzip'             => 'application/vnd.stepmania.package',
		'snd'               => 'audio/basic',
		'snf'               => 'application/x-font-snf',
		'so'                => 'application/octet-stream',
		'spc'               => 'application/x-pkcs7-certificates',
		'spd'               => 'application/x-font-speedo',
		'spec'              => 'text/x-rpm-spec',
		'spf'               => 'application/vnd.yamaha.smaf-phrase',
		'spl'               => 'application/x-futuresplash',
		'spm'               => 'application/x-source-rpm',
		'spot'              => 'text/vnd.in3d.spot',
		'spp'               => 'application/scvp-vp-response',
		'spq'               => 'application/scvp-vp-request',
		'spx'               => 'audio/ogg',
		'sql'               => 'application/x-sql',
		'sr2'               => 'image/x-sony-sr2',
		'src'               => 'application/x-wais-source',
		'src.rpm'           => 'application/x-source-rpm',
		'srf'               => 'image/x-sony-srf',
		'srt'               => 'application/x-subrip',
		'sru'               => 'application/sru+xml',
		'srx'               => 'application/sparql-results+xml',
		'ss'                => 'text/x-scheme',
		'ssa'               => 'text/x-ssa',
		'ssdl'              => 'application/ssdl+xml',
		'sse'               => 'application/vnd.kodak-descriptor',
		'ssf'               => 'application/vnd.epson.ssf',
		'ssml'              => 'application/ssml+xml',
		'st'                => 'application/vnd.sailingtracker.track',
		'stc'               => 'application/vnd.sun.xml.calc.template',
		'std'               => 'application/vnd.sun.xml.draw.template',
		'stf'               => 'application/vnd.wt.stf',
		'sti'               => 'application/vnd.sun.xml.impress.template',
		'stk'               => 'application/hyperstudio',
		'stl'               => 'application/vnd.ms-pki.stl',
		'stm'               => 'audio/x-stm',
		'str'               => 'application/vnd.pg.format',
		'stw'               => 'application/vnd.sun.xml.writer.template',
		'sty'               => 'text/x-tex',
		'sub'               => 'text/vnd.dvb.subtitle',
		'sun'               => 'image/x-sun-raster',
		'sus'               => 'application/vnd.sus-calendar',
		'susp'              => 'application/vnd.sus-calendar',
		'sv'                => 'text/x-svsrc',
		'sv4cpio'           => 'application/x-sv4cpio',
		'sv4crc'            => 'application/x-sv4crc',
		'svc'               => 'application/vnd.dvb.service',
		'svd'               => 'application/vnd.svd',
		'svg'               => 'image/svg+xml',
		'svgz'              => 'image/svg+xml',
		'svh'               => 'text/x-svhdr',
		'swa'               => 'application/x-director',
		'swf'               => 'application/x-shockwave-flash',
		'swi'               => 'application/vnd.aristanetworks.swi',
		'swm'               => 'application/x-ms-wim',
		'sxc'               => 'application/vnd.sun.xml.calc',
		'sxd'               => 'application/vnd.sun.xml.draw',
		'sxg'               => 'application/vnd.sun.xml.writer.global',
		'sxi'               => 'application/vnd.sun.xml.impress',
		'sxm'               => 'application/vnd.sun.xml.math',
		'sxw'               => 'application/vnd.sun.xml.writer',
		'sylk'              => 'text/spreadsheet',
		't'                 => 'text/troff',
		't2t'               => 'text/x-txt2tags',
		't3'                => 'application/x-t3vm-image',
		'taglet'            => 'application/vnd.mynfc',
		'tao'               => 'application/vnd.tao.intent-module-archive',
		'tar'               => 'application/x-tar',
		'tar.bz'            => 'application/x-bzip-compressed-tar',
		'tar.bz2'           => 'application/x-bzip-compressed-tar',
		'tar.gz'            => 'application/x-compressed-tar',
		'tar.lrz'           => 'application/x-lrzip-compressed-tar',
		'tar.lzma'          => 'application/x-lzma-compressed-tar',
		'tar.lzo'           => 'application/x-tzo',
		'tar.xz'            => 'application/x-xz-compressed-tar',
		'tar.z'             => 'application/x-tarz',
		'taz'               => 'application/x-tarz',
		'tb2'               => 'application/x-bzip-compressed-tar',
		'tbz'               => 'application/x-bzip-compressed-tar',
		'tbz2'              => 'application/x-bzip-compressed-tar',
		'tcap'              => 'application/vnd.3gpp2.tcap',
		'tcl'               => 'application/x-tcl',
		'teacher'           => 'application/vnd.smart.teacher',
		'tei'               => 'application/tei+xml',
		'teicorpus'         => 'application/tei+xml',
		'tex'               => 'application/x-tex',
		'texi'              => 'application/x-texinfo',
		'texinfo'           => 'application/x-texinfo',
		'text'              => 'text/plain',
		'tfi'               => 'application/thraud+xml',
		'tfm'               => 'application/x-tex-tfm',
		'tga'               => 'image/x-tga',
		'tgz'               => 'application/x-gtar',
		'theme'             => 'application/x-theme',
		'themepack'         => 'application/x-windows-themepack',
		'thmx'              => 'application/vnd.ms-officetheme',
		'tif'               => 'image/tiff',
		'tiff'              => 'image/tiff',
		'tk'                => 'text/x-tcl',
		'tlrz'              => 'application/x-lrzip-compressed-tar',
		'tlz'               => 'application/x-lzma-compressed-tar',
		'tmo'               => 'application/vnd.tmobile-livetv',
		'tnef'              => 'application/vnd.ms-tnef',
		'tnf'               => 'application/vnd.ms-tnef',
		'toc'               => 'application/x-cdrdao-toc',
		'torrent'           => 'application/x-bittorrent',
		'tpic'              => 'image/x-tga',
		'tpl'               => 'application/vnd.groove-tool-template',
		'tpt'               => 'application/vnd.trid.tpt',
		'tr'                => 'text/troff',
		'tra'               => 'application/vnd.trueapp',
		'trig'              => 'application/x-trig',
		'trm'               => 'application/x-msterminal',
		'ts'                => 'text/vnd.trolltech.linguist',
		'tsd'               => 'application/timestamped-data',
		'tsv'               => 'text/tab-separated-values',
		'tta'               => 'audio/x-tta',
		'ttc'               => 'application/x-font-ttf',
		'ttf'               => 'application/x-font-ttf',
		'ttl'               => 'text/turtle',
		'ttx'               => 'application/x-font-ttx',
		'twd'               => 'application/vnd.simtech-mindmapper',
		'twds'              => 'application/vnd.simtech-mindmapper',
		'txd'               => 'application/vnd.genomatix.tuxedo',
		'txf'               => 'application/vnd.mobius.txf',
		//'txt'               => 'text/plain',	// moved to first extry, so text/plain maps back to .txt
		'txz'               => 'application/x-xz-compressed-tar',
		'tzo'               => 'application/x-tzo',
		'u32'               => 'application/x-authorware-bin',
		'udeb'              => 'application/x-debian-package',
		'ufd'               => 'application/vnd.ufdl',
		'ufdl'              => 'application/vnd.ufdl',
		'ufraw'             => 'application/x-ufraw',
		'ui'                => 'application/x-designer',
		'uil'               => 'text/x-uil',
		'ult'               => 'audio/x-mod',
		'ulx'               => 'application/x-glulx',
		'umj'               => 'application/vnd.umajin',
		'uni'               => 'audio/x-mod',
		'unityweb'          => 'application/vnd.unity',
		'uoml'              => 'application/vnd.uoml+xml',
		'uri'               => 'text/uri-list',
		'uris'              => 'text/uri-list',
		'url'               => 'application/x-mswinurl',
		'urls'              => 'text/uri-list',
		'ustar'             => 'application/x-ustar',
		'utz'               => 'application/vnd.uiq.theme',
		'uu'                => 'text/x-uuencode',
		'uue'               => 'text/x-uuencode',
		'uva'               => 'audio/vnd.dece.audio',
		'uvd'               => 'application/vnd.dece.data',
		'uvf'               => 'application/vnd.dece.data',
		'uvg'               => 'image/vnd.dece.graphic',
		'uvh'               => 'video/vnd.dece.hd',
		'uvi'               => 'image/vnd.dece.graphic',
		'uvm'               => 'video/vnd.dece.mobile',
		'uvp'               => 'video/vnd.dece.pd',
		'uvs'               => 'video/vnd.dece.sd',
		'uvt'               => 'application/vnd.dece.ttml+xml',
		'uvu'               => 'video/vnd.uvvu.mp4',
		'uvv'               => 'video/vnd.dece.video',
		'uvva'              => 'audio/vnd.dece.audio',
		'uvvd'              => 'application/vnd.dece.data',
		'uvvf'              => 'application/vnd.dece.data',
		'uvvg'              => 'image/vnd.dece.graphic',
		'uvvh'              => 'video/vnd.dece.hd',
		'uvvi'              => 'image/vnd.dece.graphic',
		'uvvm'              => 'video/vnd.dece.mobile',
		'uvvp'              => 'video/vnd.dece.pd',
		'uvvs'              => 'video/vnd.dece.sd',
		'uvvt'              => 'application/vnd.dece.ttml+xml',
		'uvvu'              => 'video/vnd.uvvu.mp4',
		'uvvv'              => 'video/vnd.dece.video',
		'uvvx'              => 'application/vnd.dece.unspecified',
		'uvvz'              => 'application/vnd.dece.zip',
		'uvx'               => 'application/vnd.dece.unspecified',
		'uvz'               => 'application/vnd.dece.zip',
		'v'                 => 'text/x-verilog',
		'vala'              => 'text/x-vala',
		'vapi'              => 'text/x-vala',
		'vcard'             => 'text/vcard',
		'vcd'               => 'application/x-cdlink',
		'vcf'               => 'text/x-vcard',
		'vcg'               => 'application/vnd.groove-vcard',
		'vcs'               => 'text/calendar',
		'vct'               => 'text/vcard',
		'vcx'               => 'application/vnd.vcx',
		'vda'               => 'image/x-tga',
		'vfb'               => 'text/calendar',
		'vhd'               => 'text/x-vhdl',
		'vhdl'              => 'text/x-vhdl',
		'vis'               => 'application/vnd.visionary',
		'viv'               => 'video/vnd.vivo',
		'vivo'              => 'video/vivo',
		'vlc'               => 'audio/x-mpegurl',
		'vob'               => 'video/x-ms-vob',
		'voc'               => 'audio/x-voc',
		'vor'               => 'application/vnd.stardivision.writer',
		'vox'               => 'application/x-authorware-bin',
		'vrm'               => 'model/vrml',
		'vrml'              => 'model/vrml',
		'vsd'               => 'application/vnd.visio',
		'vsf'               => 'application/vnd.vsf',
		'vss'               => 'application/vnd.visio',
		'vst'               => 'application/vnd.visio',
		'vsw'               => 'application/vnd.visio',
		'vtt'               => 'text/vtt',
		'vtu'               => 'model/vnd.vtu',
		'vxml'              => 'application/voicexml+xml',
		'w3d'               => 'application/x-director',
		'wad'               => 'application/x-doom',
		'wav'               => 'audio/x-wav',
		'wax'               => 'audio/x-ms-wax',
		'wb1'               => 'application/x-quattropro',
		'wb2'               => 'application/x-quattropro',
		'wb3'               => 'application/x-quattropro',
		'wbmp'              => 'image/vnd.wap.wbmp',
		'wbs'               => 'application/vnd.criticaltools.wbs+xml',
		'wbxml'             => 'application/vnd.wap.wbxml',
		'wcm'               => 'application/vnd.ms-works',
		'wdb'               => 'application/vnd.ms-works',
		'wdp'               => 'image/vnd.ms-photo',
		'weba'              => 'audio/webm',
		'webm'              => 'video/webm',
		'webp'              => 'image/webp',
		'wg'                => 'application/vnd.pmi.widget',
		'wgt'               => 'application/widget',
		'wim'               => 'application/x-ms-wim',
		'wk1'               => 'application/vnd.lotus-1-2-3',
		'wk3'               => 'application/vnd.lotus-1-2-3',
		'wk4'               => 'application/vnd.lotus-1-2-3',
		'wks'               => 'application/vnd.ms-works',
		'wm'                => 'video/x-ms-wm',
		'wma'               => 'audio/x-ms-wma',
		'wmd'               => 'application/x-ms-wmd',
		'wmf'               => 'application/x-msmetafile',
		'wml'               => 'text/vnd.wap.wml',
		'wmlc'              => 'application/vnd.wap.wmlc',
		'wmls'              => 'text/vnd.wap.wmlscript',
		'wmlsc'             => 'application/vnd.wap.wmlscriptc',
		'wmv'               => 'video/x-ms-wmv',
		'wmx'               => 'video/x-ms-wmx',
		'wmz'               => 'application/x-ms-wmz',
		'woff'              => 'application/font-woff',
		'wp'                => 'application/vnd.wordperfect',
		'wp4'               => 'application/vnd.wordperfect',
		'wp5'               => 'application/vnd.wordperfect',
		'wp6'               => 'application/vnd.wordperfect',
		'wpd'               => 'application/vnd.wordperfect',
		'wpg'               => 'application/x-wpg',
		'wpl'               => 'application/vnd.ms-wpl',
		'wpp'               => 'application/vnd.wordperfect',
		'wps'               => 'application/vnd.ms-works',
		'wqd'               => 'application/vnd.wqd',
		'wri'               => 'application/x-mswrite',
		'wrl'               => 'model/vrml',
		'wsdl'              => 'application/wsdl+xml',
		'wsgi'              => 'text/x-python',
		'wspolicy'          => 'application/wspolicy+xml',
		'wtb'               => 'application/vnd.webturbo',
		'wv'                => 'audio/x-wavpack',
		'wvc'               => 'audio/x-wavpack-correction',
		'wvp'               => 'audio/x-wavpack',
		'wvx'               => 'video/x-ms-wvx',
		'wwf'               => 'application/x-wwf',
		'x32'               => 'application/x-authorware-bin',
		'x3d'               => 'model/x3d+xml',
		'x3db'              => 'model/x3d+binary',
		'x3dbz'             => 'model/x3d+binary',
		'x3dv'              => 'model/x3d+vrml',
		'x3dvz'             => 'model/x3d+vrml',
		'x3dz'              => 'model/x3d+xml',
		'x3f'               => 'image/x-sigma-x3f',
		'xac'               => 'application/x-gnucash',
		'xaml'              => 'application/xaml+xml',
		'xap'               => 'application/x-silverlight-app',
		'xar'               => 'application/vnd.xara',
		'xbap'              => 'application/x-ms-xbap',
		'xbd'               => 'application/vnd.fujixerox.docuworks.binder',
		'xbel'              => 'application/x-xbel',
		'xbl'               => 'application/xml',
		'xbm'               => 'image/x-xbitmap',
		'xcf'               => 'image/x-xcf',
		'xcf.bz2'           => 'image/x-compressed-xcf',
		'xcf.gz'            => 'image/x-compressed-xcf',
		'xdf'               => 'application/xcap-diff+xml',
		'xdm'               => 'application/vnd.syncml.dm+xml',
		'xdp'               => 'application/vnd.adobe.xdp+xml',
		'xdssc'             => 'application/dssc+xml',
		'xdw'               => 'application/vnd.fujixerox.docuworks',
		'xenc'              => 'application/xenc+xml',
		'xer'               => 'application/patch-ops-error+xml',
		'xfdf'              => 'application/vnd.adobe.xfdf',
		'xfdl'              => 'application/vnd.xfdl',
		'xht'               => 'application/xhtml+xml',
		'xhtml'             => 'application/xhtml+xml',
		'xhvml'             => 'application/xv+xml',
		'xi'                => 'audio/x-xi',
		'xif'               => 'image/vnd.xiff',
		'xla'               => 'application/vnd.ms-excel',
		'xlam'              => 'application/vnd.ms-excel.addin.macroenabled.12',
		'xlc'               => 'application/vnd.ms-excel',
		'xld'               => 'application/vnd.ms-excel',
		'xlf'               => 'application/x-xliff+xml',
		'xliff'             => 'application/x-xliff',
		'xll'               => 'application/vnd.ms-excel',
		'xlm'               => 'application/vnd.ms-excel',
		'xlr'               => 'application/vnd.ms-works',
		'xls'               => 'application/vnd.ms-excel',
		'xlsb'              => 'application/vnd.ms-excel.sheet.binary.macroenabled.12',
		'xlsm'              => 'application/vnd.ms-excel.sheet.macroenabled.12',
		'xlsx'              => 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
		'xlt'               => 'application/vnd.ms-excel',
		'xltm'              => 'application/vnd.ms-excel.template.macroenabled.12',
		'xltx'              => 'application/vnd.openxmlformats-officedocument.spreadsheetml.template',
		'xlw'               => 'application/vnd.ms-excel',
		'xm'                => 'audio/xm',
		'xmf'               => 'audio/x-xmf',
		'xmi'               => 'text/x-xmi',
		'xml'               => 'application/xml',
		'xo'                => 'application/vnd.olpc-sugar',
		'xop'               => 'application/xop+xml',
		'xpi'               => 'application/x-xpinstall',
		'xpl'               => 'application/xproc+xml',
		'xpm'               => 'image/x-xpixmap',
		'xpr'               => 'application/vnd.is-xpr',
		'xps'               => 'application/vnd.ms-xpsdocument',
		'xpw'               => 'application/vnd.intercon.formnet',
		'xpx'               => 'application/vnd.intercon.formnet',
		'xsd'               => 'application/xml',
		'xsl'               => 'application/xml',
		'xslfo'             => 'text/x-xslfo',
		'xslt'              => 'application/xslt+xml',
		'xsm'               => 'application/vnd.syncml+xml',
		'xspf'              => 'application/xspf+xml',
		'xul'               => 'application/vnd.mozilla.xul+xml',
		'xvm'               => 'application/xv+xml',
		'xvml'              => 'application/xv+xml',
		'xwd'               => 'image/x-xwindowdump',
		'xyz'               => 'chemical/x-xyz',
		'xz'                => 'application/x-xz',
		'yaml'              => 'application/x-yaml',
		'yang'              => 'application/yang',
		'yin'               => 'application/yin+xml',
		'yml'               => 'application/x-yaml',
		'z'                 => 'application/x-compress',
		'z1'                => 'application/x-zmachine',
		'z2'                => 'application/x-zmachine',
		'z3'                => 'application/x-zmachine',
		'z4'                => 'application/x-zmachine',
		'z5'                => 'application/x-zmachine',
		'z6'                => 'application/x-zmachine',
		'z7'                => 'application/x-zmachine',
		'z8'                => 'application/x-zmachine',
		'zabw'              => 'application/x-abiword',
		'zaz'               => 'application/vnd.zzazz.deck+xml',
		'zip'               => 'application/zip',
		'zir'               => 'application/vnd.zul',
		'zirz'              => 'application/vnd.zul',
		'zmm'               => 'application/vnd.handheld-entertainment+xml',
		'zoo'               => 'application/x-zoo',
		'zsav'              => 'application/x-spss-sav',
		'123'               => 'application/vnd.lotus-1-2-3',
		'602'               => 'application/x-t602',
		'669'               => 'audio/x-mod',
		// EGroupware own xet files
		'xet'   => 'application/x-egroupware-etemplate',
	);
	/**
	 * Mapping some old formats to the newer form, or any other aliasing for mime-types
	 *
	 * Should be in sync with ../js/jsapi/egw_image.js
	 */
	static $mime_alias_map = array(
		'text/vcard' => 'text/x-vcard',
		'text/comma-separated-values' => 'text/csv',
		'text/rtf' => 'application/rtf',
		'text/xml' => 'application/xml',
		'text/x-diff' => 'text/diff',
		'application/x-jar' => 'application/java-archive',
		'application/x-javascript' => 'application/javascript',
		'application/x-troff' => 'text/troff',
		'application/x-egroupware-etemplate' => 'application/xml',
		'image/pdf' => 'application/pdf',
	);

	/**
	 * Get the mime magic mapping file - last resort test
	 *
	 * Note Taken from horde.org - no copyright notice attached
	 *
	 * @author skwashd - converted to a function
	 * @author ralfbecker - convert to a static array ;-)
	 *
	 * @var array mime magic data
	 */
	static $mime_magic_file = array(
		0 => array(
			30 => array(
				"\145\166\141\154\40\42\145\170\145\143\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
			),
			24 => array(
				"\145\166\141\154\40\42\145\170\145\143\40\57\165\163\162\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
			),
			23 => array(
				"\103\157\155\155\157\156\40\163\165\142\144\151\162\145\143\164\157\162\151\145\163\72\40" => 'text/x-patch',
				"\75\74\154\151\163\164\76\156\74\160\162\157\164\157\143\157\154\40\142\142\156\55\155" => 'application/data',
			),
			22 => array(
				"\101\115\101\116\104\101\72\40\124\101\120\105\123\124\101\122\124\40\104\101\124\105" => 'application/x-amanda-header',
				"\107\106\61\120\101\124\103\110\61\60\60\60\111\104\43\60\60\60\60\60\62\60" => 'audio/x-gus-patch',
				"\107\106\61\120\101\124\103\110\61\61\60\60\111\104\43\60\60\60\60\60\62\60" => 'audio/x-gus-patch',
				"\43\41\11\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\142\141\163\150" => 'application/x-sh',
				"\43\41\11\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\11\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\11\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\43\41\11\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\164\143\163\150" => 'application/x-csh',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\142\141\163\150" => 'application/x-sh',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\164\143\163\150" => 'application/x-csh',
			),
			21 => array(
				"\43\41\11\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\141\163\150" => 'application/x-zsh',
				"\43\41\11\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\172\163\150" => 'application/x-zsh',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\141\163\150" => 'application/x-zsh',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\172\163\150" => 'application/x-zsh',
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\142\141\163\150" => 'application/x-sh',
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\164\143\163\150" => 'application/x-csh',
			),
			20 => array(
				"\145\166\141\154\40\42\145\170\145\143\40\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\43\41\11\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\141\145" => 'text/script',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\141\145" => 'text/script',
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\141\163\150" => 'application/x-sh',
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\172\163\150" => 'application/x-zsh',
			),
			19 => array(
				"\103\162\145\141\164\151\166\145\40\126\157\151\143\145\40\106\151\154\145" => 'audio/x-voc',
				"\41\74\141\162\143\150\76\156\137\137\137\137\137\137\137\137\137\137\105" => 'application/x-ar',
				"\41\74\141\162\143\150\76\156\137\137\137\137\137\137\137\137\66\64\105" => 'application/data',
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\142\151\156\57\141\145" => 'text/script',
			),
			18 => array(
				"\106\151\114\145\123\164\101\162\124\146\111\154\105\163\124\141\122\164" => 'text/x-apple-binscii',
				"\43\41\40\57\165\163\162\57\154\157\143\141\154\57\164\143\163\150" => 'application/x-csh',
				"\45\41\120\123\55\101\144\157\142\145\106\157\156\164\55\61\56\60" => 'font/type1',
			),
			17 => array(
				"\43\41\57\165\163\162\57\154\157\143\141\154\57\164\143\163\150" => 'application/x-csh',
			),
			16 => array(
				"\105\170\164\145\156\144\145\144\40\115\157\144\165\154\145\72" => 'audio/x-ft2-mod',
				"\123\164\141\162\164\106\157\156\164\115\145\164\162\151\143\163" => 'font/x-sunos-news',
				"\43\41\11\57\165\163\162\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\11\57\165\163\162\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\11\57\165\163\162\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\43\41\40\57\165\163\162\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\40\57\165\163\162\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\40\57\165\163\162\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\74\115\141\153\145\162\104\151\143\164\151\157\156\141\162\171" => 'application/x-framemaker',
				"\74\115\141\153\145\162\123\143\162\145\145\156\106\157\156\164" => 'font/x-framemaker',
			),
			15 => array(
				"\43\41\11\57\165\163\162\57\142\151\156\57\141\167\153" => 'application/x-awk',
				"\43\41\40\57\165\163\162\57\142\151\156\57\141\167\153" => 'application/x-awk',
				"\43\41\57\165\163\162\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\57\165\163\162\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\57\165\163\162\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
			),
			14 => array(
				"\41\74\141\162\143\150\76\156\144\145\142\151\141\156" => 'application/x-dpkg',
				"\43\41\57\165\163\162\57\142\151\156\57\141\167\153" => 'application/x-awk',
				"\74\41\104\117\103\124\131\120\105\40\110\124\115\114" => 'text/html',
				"\74\41\144\157\143\164\171\160\145\40\150\164\155\154" => 'text/html',
			),
			13 => array(
				"\107\111\115\120\40\107\162\141\144\151\145\156\164" => 'application/x-gimp-gradient',
			),
			12 => array(
				"\122\145\164\165\162\156\55\120\141\164\150\72" => 'message/rfc822',
				"\43\41\11\57\142\151\156\57\142\141\163\150" => 'application/x-sh',
				"\43\41\11\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\11\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\11\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\43\41\11\57\142\151\156\57\164\143\163\150" => 'application/x-csh',
				"\43\41\40\57\142\151\156\57\142\141\163\150" => 'application/x-sh',
				"\43\41\40\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\40\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\40\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\43\41\40\57\142\151\156\57\164\143\163\150" => 'application/x-csh',
			),
			11 => array(
				"\43\41\11\57\142\151\156\57\141\167\153" => 'application/x-awk',
				"\43\41\11\57\142\151\156\57\143\163\150" => 'application/x-csh',
				"\43\41\11\57\142\151\156\57\153\163\150" => 'application/x-ksh',
				"\43\41\40\57\142\151\156\57\141\167\153" => 'application/x-awk',
				"\43\41\40\57\142\151\156\57\143\163\150" => 'application/x-csh',
				"\43\41\40\57\142\151\156\57\153\163\150" => 'application/x-ksh',
				"\43\41\57\142\151\156\57\142\141\163\150" => 'application/x-sh',
				"\43\41\57\142\151\156\57\147\141\167\153" => 'application/x-awk',
				"\43\41\57\142\151\156\57\156\141\167\153" => 'application/x-awk',
				"\43\41\57\142\151\156\57\160\145\162\154" => 'application/x-perl',
				"\43\41\57\142\151\156\57\164\143\163\150" => 'application/x-csh',
			),
			10 => array(
				"\102\151\164\155\141\160\146\151\154\145" => 'image/unknown',
				"\123\124\101\122\124\106\117\116\124\40" => 'font/x-bdf',
				"\43\41\11\57\142\151\156\57\162\143" => 'text/script',
				"\43\41\11\57\142\151\156\57\163\150" => 'application/x-sh',
				"\43\41\40\57\142\151\156\57\162\143" => 'text/script',
				"\43\41\40\57\142\151\156\57\163\150" => 'application/x-sh',
				"\43\41\57\142\151\156\57\141\167\153" => 'application/x-awk',
				"\43\41\57\142\151\156\57\143\163\150" => 'application/x-csh',
				"\43\41\57\142\151\156\57\153\163\150" => 'application/x-ksh',
				"\74\115\141\153\145\162\106\151\154\145" => 'application/x-framemaker',
			),
			9 => array(
				"\122\145\143\145\151\166\145\144\72" => 'message/rfc822',
				"\123\164\141\162\164\106\157\156\164" => 'font/x-sunos-news',
				"\211\114\132\117\0\15\12\32\12" => 'application/data',
				"\43\41\57\142\151\156\57\162\143" => 'text/script',
				"\43\41\57\142\151\156\57\163\150" => 'application/x-sh',
				"\55\162\157\155\61\146\163\55\60" => 'application/x-filesystem',
				"\74\102\157\157\153\106\151\154\145" => 'application/x-framemaker',
			),
			8 => array(
				"\117\156\154\171\40\151\156\40" => 'text/x-patch',
				"\147\151\155\160\40\170\143\146" => 'application/x-gimp-image',
				"\155\163\147\143\141\164\60\61" => 'application/x-locale',
				"\32\141\162\143\150\151\166\145" => 'application/data',
				"\41\74\120\104\106\76\41\156" => 'application/x-prof',
				"\74\115\111\106\106\151\154\145" => 'application/x-framemaker',
			),
			7 => array(
				"\101\162\164\151\143\154\145" => 'message/news',
				"\120\103\104\137\117\120\101" => 'x/x-photo-cd-overfiew-file',
				"\351\54\1\112\101\115\11" => 'application/data',
				"\41\74\141\162\143\150\76" => 'application/x-ar',
				"\72\40\163\150\145\154\154" => 'application/data',
			),
			6 => array(
				"\116\165\106\151\154\145" => 'application/data',
				"\116\365\106\351\154\345" => 'application/data',
				"\60\67\60\67\60\61" => 'application/x-cpio',
				"\60\67\60\67\60\62" => 'application/x-cpio',
				"\60\67\60\67\60\67" => 'application/x-cpio',
				"\74\115\141\153\145\162" => 'application/x-framemaker',
				"\74\124\111\124\114\105" => 'text/html',
				"\74\164\151\164\154\145" => 'text/html',
			),
			5 => array(
				"\0\1\0\0\0" => 'font/ttf',
				"\0\4\36\212\200" => 'application/core',
				"\102\101\102\131\114" => 'message/x-gnu-rmail',
				"\102\105\107\111\116" => 'application/x-awk',
				"\103\157\162\145\1" => 'application/x-executable-file',
				"\104\61\56\60\15" => 'font/x-speedo',
				"\106\162\157\155\72" => 'message/rfc822',
				"\115\101\123\137\125" => 'audio/x-multimate-mod',
				"\120\117\136\121\140" => 'text/vnd.ms-word',
				"\120\141\164\150\72" => 'message/news',
				"\130\162\145\146\72" => 'message/news',
				"\144\151\146\146\40" => 'text/x-patch',
				"\225\64\62\62\336" => 'application/x-locale',
				"\336\62\62\64\225" => 'application/x-locale',
				"\74\110\105\101\104" => 'text/html',
				"\74\110\124\115\114" => 'text/html',
				"\74\150\145\141\144" => 'text/html',
				"\74\150\164\155\154" => 'text/html',
				"\75\74\141\162\76" => 'application/x-ar',
			),
			4 => array(
				"\0\0\0\314" => 'application/x-executable-file',
				"\0\0\0\4" => 'font/x-snf',
				"\0\0\1\107" => 'application/x-object-file',
				"\0\0\1\113" => 'application/x-executable-file',
				"\0\0\1\115" => 'application/x-executable-file',
				"\0\0\1\117" => 'application/x-executable-file',
				"\0\0\1\201" => 'application/x-object-file',
				"\0\0\1\207" => 'application/data',
				"\0\0\1\263" => 'video/mpeg',
				"\0\0\1\272" => 'video/mpeg',
				"\0\0\1\6" => 'application/x-executable-file',
				"\0\0\201\154" => 'application/x-apl-workspace',
				"\0\0\377\145" => 'application/x-library-file',
				"\0\0\377\155" => 'application/data',
				"\0\0\3\347" => 'application/x-library-file',
				"\0\0\3\363" => 'application/x-executable-file',
				"\0\144\163\56" => 'audio/basic',
				"\0\1\22\127" => 'application/core',
				"\0\22\326\207" => 'image/x11',
				"\0\3\233\355" => 'application/data',
				"\0\3\233\356" => 'application/data',
				"\0\5\26\0" => 'application/data',
				"\0\5\26\7" => 'application/data',
				"\0\5\61\142" => 'application/x-db',
				"\0\6\25\141" => 'application/x-db',
				"\103\124\115\106" => 'audio/x-cmf',
				"\105\115\117\104" => 'audio/x-emod',
				"\106\106\111\114" => 'font/ttf',
				"\106\117\116\124" => 'font/x-vfont',
				"\107\104\102\115" => 'application/x-gdbm',
				"\107\111\106\70" => 'image/gif',
				"\10\16\12\17" => 'application/data',
				"\110\120\101\113" => 'application/data',
				"\111\111\116\61" => 'image/tiff',
				"\111\111\52\0" => 'image/tiff',
				"\114\104\110\151" => 'application/data',
				"\114\127\106\116" => 'font/type1',
				"\115\115\0\52" => 'image/tiff',
				"\115\117\126\111" => 'video/x-sgi-movie',
				"\115\124\150\144" => 'audio/midi',
				"\115\247\356\350" => 'font/x-hp-windows',
				"\116\124\122\113" => 'audio/x-multitrack',
				"\120\113\3\4" => 'application/zip',
				"\122\111\106\106" => 'audio/x-wav',
				"\122\141\162\41" => 'application/x-rar',
				"\123\121\123\110" => 'application/data',
				"\124\101\104\123" => 'application/x-tads-game',
				"\125\103\62\32" => 'application/data',
				"\125\116\60\65" => 'audio/x-mikmod-uni',
				"\12\17\10\16" => 'application/data',
				"\131\246\152\225" => 'x/x-image-sun-raster',
				"\145\377\0\0" => 'application/x-ar',
				"\150\163\151\61" => 'image/x-jpeg-proprietary',
				"\16\10\17\12" => 'application/data',
				"\177\105\114\106" => 'application/x-executable-file',
				"\17\12\16\10" => 'application/data',
				"\1\130\41\246" => 'application/core',
				"\1\146\143\160" => 'font/x-pcf',
				"\211\120\116\107" => 'image/x-png',
				"\23\127\232\316" => 'application/x-gdbm',
				"\23\172\51\104" => 'font/x-sunos-news',
				"\23\172\51\107" => 'font/x-sunos-news',
				"\23\172\51\120" => 'font/x-sunos-news',
				"\23\172\51\121" => 'font/x-sunos-news',
				"\24\2\131\31" => 'font/x-libgrx',
				"\260\61\63\140" => 'application/x-bootable',
				"\2\10\1\10" => 'application/x-executable-file',
				"\2\10\1\6" => 'application/x-executable-file',
				"\2\10\1\7" => 'application/x-executable-file',
				"\2\10\377\145" => 'application/x-library-file',
				"\2\12\1\10" => 'application/x-executable-file',
				"\2\12\1\7" => 'application/x-executable-file',
				"\2\12\377\145" => 'application/x-library-file',
				"\2\13\1\10" => 'application/x-executable-file',
				"\2\13\1\13" => 'application/x-executable-file',
				"\2\13\1\15" => 'application/x-library-file',
				"\2\13\1\16" => 'application/x-library-file',
				"\2\13\1\6" => 'application/x-object-file',
				"\2\13\1\7" => 'application/x-executable-file',
				"\2\14\1\10" => 'application/x-executable-file',
				"\2\14\1\13" => 'application/x-executable-file',
				"\2\14\1\14" => 'application/x-lisp',
				"\2\14\1\15" => 'application/x-library-file',
				"\2\14\1\16" => 'application/x-library-file',
				"\2\14\1\6" => 'application/x-executable-file',
				"\2\14\1\7" => 'application/x-executable-file',
				"\2\14\377\145" => 'application/x-library-file',
				"\2\20\1\10" => 'application/x-executable-file',
				"\2\20\1\13" => 'application/x-executable-file',
				"\2\20\1\15" => 'application/x-library-file',
				"\2\20\1\16" => 'application/x-library-file',
				"\2\20\1\6" => 'application/x-object-file',
				"\2\20\1\7" => 'application/x-executable-file',
				"\2\24\1\10" => 'application/x-executable-file',
				"\2\24\1\13" => 'application/x-executable-file',
				"\2\24\1\15" => 'application/x-object-file',
				"\2\24\1\16" => 'application/x-library-file',
				"\2\24\1\6" => 'application/x-object-file',
				"\2\24\1\7" => 'application/x-executable-file',
				"\361\60\100\273" => 'image/x-cmu-raster',
				"\366\366\366\366" => 'application/x-pc-floppy',
				"\377\106\117\116" => 'font/x-dos',
				"\41\74\141\162" => 'application/x-ar',
				"\43\41\11\57" => 'text/script',
				"\43\41\40\57" => 'text/script',
				"\52\123\124\101" => 'application/data',
				"\52\52\52\40" => 'text/x-patch',
				"\56\162\141\375" => 'audio/x-pn-realaudio',
				"\56\163\156\144" => 'audio/basic',
				"\61\143\167\40" => 'application/data',
				"\61\276\0\0" => 'text/vnd.ms-word',
				"\62\62\67\70" => 'application/data',
				"\74\115\115\114" => 'application/x-framemaker',
				"\74\141\162\76" => 'application/x-ar',
			),
			3 => array(
				"\102\132\150" => 'application/x-bzip2',
				"\106\101\122" => 'audio/mod',
				"\115\124\115" => 'audio/x-multitrack',
				"\123\102\111" => 'audio/x-sbi',
				"\124\117\103" => 'audio/x-toc',
				"\12\107\114" => 'application/data',
				"\146\154\143" => 'application/x-font',
				"\146\154\146" => 'font/x-figlet',
				"\33\105\33" => 'image/x-pcl-hp',
				"\33\143\33" => 'application/data',
				"\377\377\174" => 'application/data',
				"\377\377\176" => 'application/data',
				"\377\377\177" => 'application/data',
				"\43\41\40" => 'text/script',
				"\43\41\57" => 'text/script',
				"\4\45\41" => 'application/postscript',
				"\55\150\55" => 'application/data',
				"\61\143\167" => 'application/data',
			),
			2 => array(
				"\0\0" => 'application/x-executable-file',
				"\102\115" => 'image/x-bmp',
				"\102\132" => 'application/x-bzip',
				"\111\103" => 'image/x-ico',
				"\112\116" => 'audio/x-669-mod',
				"\115\132" => 'application/x-ms-dos-executable',
				"\120\61" => 'image/x-portable-bitmap',
				"\120\62" => 'image/x-portable-graymap',
				"\120\63" => 'image/x-portable-pixmap',
				"\120\64" => 'image/x-portable-bitmap',
				"\120\65" => 'image/x-portable-graymap',
				"\120\66" => 'image/x-portable-pixmap',
				"\151\146" => 'audio/x-669-mod',
				"\161\307" => 'application/x-cpio',
				"\166\377" => 'application/data',
				"\1\110" => 'application/x-executable-file',
				"\1\111" => 'application/x-executable-file',
				"\1\124" => 'application/data',
				"\1\125" => 'application/x-executable-file',
				"\1\160" => 'application/x-executable-file',
				"\1\161" => 'application/x-executable-file',
				"\1\175" => 'application/x-executable-file',
				"\1\177" => 'application/x-executable-file',
				"\1\20" => 'application/x-executable-file',
				"\1\203" => 'application/x-executable-file',
				"\1\21" => 'application/x-executable-file',
				"\1\210" => 'application/x-executable-file',
				"\1\217" => 'application/x-object-file',
				"\1\224" => 'application/x-executable-file',
				"\1\227" => 'application/x-executable-file',
				"\1\332" => 'x/x-image-sgi',
				"\1\36" => 'font/x-vfont',
				"\1\6" => 'application/x-executable-file',
				"\307\161" => 'application/x-bcpio',
				"\313\5" => 'application/data',
				"\352\140" => 'application/x-arj',
				"\367\131" => 'font/x-tex',
				"\367\203" => 'font/x-tex',
				"\367\312" => 'font/x-tex',
				"\36\1" => 'font/x-vfont',
				"\375\166" => 'application/x-lzh',
				"\376\166" => 'application/data',
				"\377\145" => 'application/data',
				"\377\155" => 'application/data',
				"\377\166" => 'application/data',
				"\377\330" => 'image/jpeg',
				"\377\37" => 'application/data',
				"\37\213" => 'application/x-gzip',
				"\37\235" => 'application/compress',
				"\37\236" => 'application/data',
				"\37\237" => 'application/data',
				"\37\240" => 'application/data',
				"\37\36" => 'application/data',
				"\37\37" => 'application/data',
				"\37\377" => 'application/data',
				"\45\41" => 'application/postscript',
				"\4\66" => 'font/linux-psf',
				"\57\57" => 'text/cpp',
				"\5\1" => 'application/x-locale',
				"\6\1" => 'application/x-executable-file',
				"\6\2" => 'application/x-alan-adventure-game',
				"\7\1" => 'application/x-executable-file',
			),
		),
		1 => array(
			3 => array(
				"\120\116\107" => 'image/x-png',
				"\127\120\103" => 'text/vnd.wordperfect',
			),
		),
		2 => array(
			6 => array(
				"\55\154\150\64\60\55" => 'application/x-lha',
			),
			5 => array(
				"\55\154\150\144\55" => 'application/x-lha',
				"\55\154\150\60\55" => 'application/x-lha',
				"\55\154\150\61\55" => 'application/x-lha',
				"\55\154\150\62\55" => 'application/x-lha',
				"\55\154\150\63\55" => 'application/x-lha',
				"\55\154\150\64\55" => 'application/x-lha',
				"\55\154\150\65\55" => 'application/x-lha',
				"\55\154\172\163\55" => 'application/x-lha',
				"\55\154\172\64\55" => 'application/x-lha',
				"\55\154\172\65\55" => 'application/x-lha',
			),
			2 => array(
				"\0\21" => 'font/x-tex-tfm',
				"\0\22" => 'font/x-tex-tfm',
			),
		),
		4 => array(
			4 => array(
				"\155\144\141\164" => 'video/quicktime',
				"\155\157\157\166" => 'video/quicktime',
				"\160\151\160\145" => 'application/data',
				"\160\162\157\146" => 'application/data',
			),
			2 => array(
				"\257\21" => 'video/fli',
				"\257\22" => 'video/flc',
			),
		),
		6 => array(
			18 => array(
				"\45\41\120\123\55\101\144\157\142\145\106\157\156\164\55\61\56\60" => 'font/type1',
			),
		),
		7 => array(
			22 => array(
				"\357\20\60\60\60\60\60\60\60\60\60\60\60\60\60\60\60\60\60\60\60\60" => 'application/core',
			),
			4 => array(
				"\0\105\107\101" => 'font/x-dos',
				"\0\126\111\104" => 'font/x-dos',
			),
		),
		8 => array(
			4 => array(
				"\23\172\53\105" => 'font/x-sunos-news',
				"\23\172\53\110" => 'font/x-sunos-news',
			),
		),
		10 => array(
			25 => array(
				"\43\40\124\150\151\163\40\151\163\40\141\40\163\150\145\154\154\40\141\162\143\150\151\166\145" => 'application/x-shar',
			),
		),
		20 => array(
			4 => array(
				"\107\111\115\120" => 'application/x-gimp-brush',
				"\107\120\101\124" => 'application/x-gimp-pattern',
				"\375\304\247\334" => 'application/x-zoo',
			),
		),
		21 => array(
			8 => array(
			"\41\123\103\122\105\101\115\41" => 'audio/x-st2-mod',
			),
		),
		24 => array(
			4 => array(
				"\0\0\352\153" => 'application/x-dump',
				"\0\0\352\154" => 'application/x-dump',
				"\0\0\352\155" => 'application/data',
				"\0\0\352\156" => 'application/data',
			),
		),
		65 => array(
			4 => array(
				"\106\106\111\114" => 'font/ttf',
				"\114\127\106\116" => 'font/type1',
			),
		),
		257 => array(
			8 => array(
				"\165\163\164\141\162\40\40\60" => 'application/x-gtar',
			),
			6 => array(
				"\165\163\164\141\162\60" => 'application/x-tar',
			),
		),
		0774 => array(
			2 => array(
				"\332\276" => 'application/data',
			),
		),
		1080 => array(
			4 => array(
				"\103\104\70\61" => 'audio/x-oktalyzer-mod',
				"\106\114\124\64" => 'audio/x-startracker-mod',
				"\115\41\113\41" => 'audio/x-protracker-mod',
				"\115\56\113\56" => 'audio/x-protracker-mod',
				"\117\113\124\101" => 'audio/x-oktalyzer-mod',
				"\61\66\103\116" => 'audio/x-taketracker-mod',
				"\63\62\103\116" => 'audio/x-taketracker-mod',
				"\64\103\110\116" => 'audio/x-fasttracker-mod',
				"\66\103\110\116" => 'audio/x-fasttracker-mod',
				"\70\103\110\116" => 'audio/x-fasttracker-mod',
			),
		),
		2048 => array(
			7 => array(
				"\120\103\104\137\111\120\111" => 'x/x-photo-cd-pack-file',
			),
		),
		2080 => array(
			29 => array(
				"\115\151\143\162\157\163\157\146\164\40\105\170\143\145\154\40\65\56\60\40\127\157\162\153\163\150\145\145\164" => 'application/vnd.ms-excel',
			),
			27 => array(
				"\115\151\143\162\157\163\157\146\164\40\127\157\162\144\40\66\56\60\40\104\157\143\165\155\145\156\164" => 'text/vnd.ms-word',
			),
			26 => array(
				"\104\157\143\165\155\145\156\164\157\40\115\151\143\162\157\163\157\146\164\40\127\157\162\144\40\66" => 'text/vnd.ms-word',
			),
		),
		2112 => array(
			9 => array(
				"\115\123\127\157\162\144\104\157\143" => 'text/vnd.ms-word',
			),
		),
		2114 => array(
			5 => array(
				"\102\151\146\146\65" => 'application/vnd.ms-excel',
			),
		),
		4098 => array(
			7 => array(
				"\104\117\123\106\117\116\124" => 'font/x-dos',
			),
		),
		68158480 => array(
			2 => array(
				"\23\177" => 'application/x-filesystem',
				"\23\217" => 'application/x-filesystem',
				"\44\150" => 'application/x-filesystem',
				"\44\170" => 'application/x-filesystem',
			),
		),
		70779960 => array(
			2 => array(
				"\357\123" => 'application/x-linux-ext2fs',
			),
		),
	);
}